from typing import List, Dict, Optional, Type, Tuple, Union, TYPE_CHECKING
from copy import deepcopy

import pandas as pd
import asyncio

from shimoku.api.base_resource import Resource, ResourceCache
from shimoku.api.resources.activity import Activity
from shimoku.api.resources.role import (
    Role,
    create_role,
    get_role,
    get_roles,
    delete_role,
)
from shimoku.api.resources.report import Report
from shimoku.api.resources.file import File
from shimoku.api.resources.data_set import DataSet, Mapping

from shimoku.utils import create_normalized_name
from shimoku.exceptions import MenuPathError, UniverseApiKeyError, ActivityTemplateError

if TYPE_CHECKING:
    from .business import Business

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


class App(Resource):
    """App resource class"""

    resource_type = "app"
    alias_field = "name"
    plural = "apps"

    def __init__(
        self,
        parent: "Business",
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[Dict] = None,
    ):
        normalized_name: Optional[str] = None
        if alias:
            normalized_name: str = create_normalized_name(alias)

        params = dict(
            name=alias,
            order=0,
            normalizedName=normalized_name,
            hideTitle=True,
            hidePath=False,
            showBreadcrumb=False,
            showHistoryNavigation=False,
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            children=[Role, Activity, Report, DataSet, File],
            check_params_before_creation=["name"],
            params=params,
        )

        self.currently_in_use: bool = False

    async def delete(self):
        """Delete the app"""
        if await self.check_for_shimoku_generated_files():
            log_error(
                logger,
                f"The menu path ({str(self)}) cannot be deleted, "
                f"because it contains files generated by the SDK. If you are sure you want to delete those files, "
                f"please use the method (menu_paths.delete_all_menu_path_files) "
                f"with the parameter (with_shimoku_generated) set to True",
                MenuPathError,
            )
        if self.currently_in_use:
            log_error(
                logger,
                f"Menu path ({str(self)}) is currently in use and cannot be deleted",
                MenuPathError,
            )
        return await self._base_resource.delete()

    async def update(self):
        """Update the app"""
        return await self._base_resource.update()

    # Activity methods
    async def _get_activity_template_id(
        self,
        template_id: Optional[str] = None,
        template_name_version: Optional[Tuple[str, str]] = None,
    ) -> str:
        """
        Get the activity template id from the template id or name
        :param template_id: the template id
        :param template_name_version: the template name and version
        """
        activity_template = await self.parent.parent.get_activity_template(
            uuid=template_id, name_version=template_name_version
        )
        if activity_template:
            return activity_template["id"]

        log_error(
            logger,
            "Activity template not found in the universe, please use an existing one.",
            ActivityTemplateError,
        )

    async def create_activity(
        self,
        name: str,
        settings: Optional[Dict[str, str]] = None,
        template_id: Optional[str] = None,
        template_name_version: Optional[Tuple[str, str]] = None,
        template_mode: str = "LIGHT",
        universe_api_key: str = "",
    ) -> Activity:
        """
        Create an activity, either from a template or from scratch
        :param name: Name of the activity
        :param settings: Settings of the activity
        :param template_id: UUID of the activity template
        :param template_name_version: Name and version of the activity template
        :param template_mode: Mode of the activity template
        :param universe_api_key: Universe API key to use
        """
        template_params_to_send = {}

        if template_id or template_name_version:
            template_id = await self._get_activity_template_id(
                template_id, template_name_version
            )
            universe_api_keys = await self.parent.parent.get_universe_api_keys()
            if universe_api_key not in [u["id"] for u in universe_api_keys]:
                log_error(
                    logger,
                    "Universe API key not found in the universe, please use an existing one.",
                    UniverseApiKeyError,
                )

            template_params_to_send = dict(
                activityTemplateWithMode=dict(
                    activityTemplateId=template_id,
                    mode=template_mode,
                ),
                universeApiKeyId=universe_api_key,
            )

        return await self._base_resource.create_child(
            Activity, alias=name, settings=settings, **template_params_to_send
        )

    async def update_activity(
        self, uuid: Optional[str] = None, name: Optional[str] = None, **params
    ) -> Activity:
        if params.get("new_name") is not None:
            params["name"] = params.pop("new_name")
            params["new_alias"] = True
        return await self._base_resource.update_child(
            Activity, uuid=uuid, alias=name, **params
        )

    async def get_activity(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[Activity]:
        return await self._base_resource.get_child(Activity, uuid, name)

    async def get_activities(self) -> List[Activity]:
        return await self._base_resource.get_children(Activity)

    async def delete_activity(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        return await self._base_resource.delete_child(Activity, uuid, name)

    # Report methods
    def mock_create_report(
        self, report_class: Type[Report], r_hash: str, **params
    ) -> Report:
        """Creates a child of a given report class but doesn't add it to the cache.
        :param report_class: The class of the report to create.
        :param r_hash: The hash of the report to create.
        :param params: The parameters of the report to create.
        :return: The created resource.
        """
        report = report_class(parent=self)
        params = deepcopy(params)

        properties = params.pop("properties") if "properties" in params else {}
        properties["hash"] = r_hash

        report.set_properties(**properties)
        report.set_params(**params)
        return report

    async def get_paths_in_order(self) -> List[str]:
        """Gets all the paths of the app. They are stored in the reports."""
        reports = await self.get_reports()
        path_orders = [
            (report["path"], report["pathOrder"])
            for report in reports
            if report["pathOrder"] is not None
        ]
        path_orders.sort(key=lambda x: x[1])
        result = []
        for path_order in path_orders:
            if path_order[0] not in result:
                result.append(path_order[0])
        return result

    async def create_report(
        self, report_class: Type[Report], r_hash: str, **params
    ) -> Report:
        """Creates a child of a given report class.
        :param report_class: The class of the report to create.
        :param r_hash: The hash of the report to create.
        :param params: The parameters of the report to create.
        :return: The created resource.
        """
        report_cache: ResourceCache = self._base_resource.children[Report]
        report = await report_cache.add(
            self.mock_create_report(report_class, r_hash, **params), alias=r_hash
        )
        return report

    async def get_report(
        self, uuid: Optional[str] = None, r_hash: Optional[str] = None
    ) -> Optional[Report]:
        """Gets a report.
        :param uuid: The UUID of the report to get.
        :param r_hash: The hash of the report to get.
        :return: The resource.
        """
        return await self._base_resource.get_child(Report, uuid, r_hash)

    async def get_reports(self) -> List[Report]:
        """Gets all the reports of the app."""
        return await self._base_resource.get_children(Report)

    async def update_report(
        self, uuid: Optional[str] = None, r_hash: Optional[str] = None, **params
    ):
        """Updates a report.
        :param uuid: The UUID of the report to update.
        :param r_hash: The hash of the report to update.
        :param params: The parameters of the report to update.
        """
        report: Optional[Report] = await self.get_report(uuid, r_hash)
        if not report:
            logger.warning(f"Report {r_hash if r_hash else uuid} not found")
            return
        params = deepcopy(params)
        if "properties" in params:
            if "hash" in params["properties"]:
                log_error(logger, "Cannot update the hash of a component.", ValueError)
            report.set_properties(**params.pop("properties"))

        report.set_params(**params)
        await report.update()

    async def delete_report(
        self, uuid: Optional[str] = None, r_hash: Optional[str] = None
    ):
        """Deletes a report.
        :param uuid: The UUID of the report to delete.
        :param r_hash: The hash of the report to delete.
        """
        report: Optional[Report] = await self.get_report(uuid, r_hash)
        if not report:
            return
        await self._base_resource.delete_child(Report, uuid, r_hash)

    # DataSet methods
    async def get_data_set(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        create_if_not_exists=True,
    ) -> Optional[DataSet]:
        """Gets a dataset.
        :param uuid: The UUID of the dataset to get.
        :param name: The alias of the dataset to get.
        :param create_if_not_exists: If True, creates the dataset if it doesn't exist.
        :return: The dataset.
        """
        return await self._base_resource.get_child(
            DataSet, uuid, name, create_if_not_exists=create_if_not_exists
        )

    async def get_data_sets(self) -> List[DataSet]:
        """Gets all the datasets of the app."""
        return await self._base_resource.get_children(DataSet)

    async def delete_data_set(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """Deletes a dataset.
        :param uuid: The UUID of the dataset to delete.
        :param name: The alias of the dataset to delete.
        """
        return await self._base_resource.delete_child(DataSet, uuid, name)

    async def append_data_to_data_set(
        self,
        data: Union[pd.DataFrame, List[Dict]],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        sort: Optional[Dict] = None,
        dump_whole: bool = False,
    ) -> Tuple[Mapping, DataSet, Dict]:
        """Appends data to a dataset. If the dataset doesn't exist, it is created.
        :param uuid: The UUID of the dataset to update.
        :param name: The alias of the dataset to update.
        :param data: The data to update the dataset with.
        :param sort: The sort to apply to the data.
        :param dump_whole: Whether to dump the whole data.
        :return: The dataset.
        """
        data_set: DataSet = await self.get_data_set(uuid, name)
        mapping, res_sort = await data_set.create_data_points(
            data_points=pd.DataFrame(data)
            if isinstance(data, list) and not dump_whole
            else data,
            sort=sort,
            dump_whole=dump_whole,
        )
        return mapping, data_set, res_sort

    async def force_delete_data_set(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        """Deletes a dataset.
        :param uuid: The UUID of the dataset to delete.
        :param name: The alias of the dataset to delete.
        :return: Whether the dataset was deleted.
        """
        data_set = await self.get_data_set(uuid, name, create_if_not_exists=False)
        if not data_set:
            return False
        reports = await self.get_reports()
        report_data_sets_lists = await asyncio.gather(
            *[report.get_report_data_sets() for report in reports]
        )
        delete_tasks = []
        for i, report_data_sets in enumerate(report_data_sets_lists):
            await asyncio.gather(
                *[
                    reports[i].delete_report_dataset(rds["id"])
                    for rds in report_data_sets
                    if rds["dataSetId"] == data_set["id"]
                ]
            )

        await asyncio.gather(*delete_tasks)
        await self.delete_data_set(uuid, name)
        return True

    async def delete_unused_data_sets(
        self, log: bool = False, data_set_ids: Optional[List[str]] = None
    ):
        """Deletes all the unused datasets of the app.
        :param log: Whether to log the deletion.
        :param data_set_ids: The ids of the datasets to delete.
        """
        all_reports = await self.get_reports()
        all_rds = await asyncio.gather(
            *[report.get_report_data_sets() for report in all_reports]
        )
        all_datasets_in_use = [
            rds["dataSetId"] for _rds_list in all_rds for rds in _rds_list
        ]
        all_datasets = await self.get_data_sets()
        dataset_ids_to_delete = [
            ds["id"]
            for ds in all_datasets
            if ds["id"] not in all_datasets_in_use
            and (ds["id"] in data_set_ids or data_set_ids is None)
        ]
        await asyncio.gather(
            *[
                self._base_resource.delete_child(DataSet, ds_id)
                for ds_id in dataset_ids_to_delete
            ]
        )
        if log:
            logger.info(
                f"Deleted {len(dataset_ids_to_delete)} unused datasets from the menu path {str(self)}"
            )

    # File methods
    async def check_for_shimoku_generated_files(self) -> bool:
        """Checks whether the app has files generated by shimoku.
        :return: Whether the app has files generated by shimoku.
        """
        files = await self.get_files()
        return any(["shimoku_generated" in file["tags"] for file in files])

    async def create_file(
        self,
        name: str,
        file_object: bytes,
        tags: list,
        metadata: dict,
        overwrite: bool = True,
    ) -> File:
        """Creates a file.
        :param name: The file to create.
        :param file_object: The file object.
        :param tags: The tags of the file.
        :param metadata: The metadata of the file.
        :param overwrite: Whether to overwrite the file if it already exists.
        :return: The created file.
        """
        if overwrite and await self._base_resource.get_child(
            File, alias=name, create_if_not_exists=False
        ):
            logger.info(f"Overwriting file {name}")
            await self.delete_file(name=name)

        name = await self._base_resource.create_child(
            File, alias=name, contentType="text/csv", tags=tags, metadata=metadata
        )
        logger.info(f"Uploading file {str(name)}")
        await self.api_client.request(
            "PUT",
            name["url"],
            body=file_object,
            headers={"content-type": "text/csv"},
            to_tazawa=False,
        )

        return name

    async def delete_file(self, uuid: Optional[str] = None, name: Optional[str] = None):
        """Deletes a file.
        :param uuid: The UUID of the file to delete.
        :param name: The name of the file to delete.
        """
        return await self._base_resource.delete_child(File, uuid, name)

    async def get_file(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[File]:
        """Gets a file.
        :param uuid: The UUID of the file to get.
        :param name: The alias of the file to get.
        :return: The file.
        """
        return await self._base_resource.get_child(File, uuid, name)

    async def get_files(self) -> List[File]:
        """Gets all the files of the app.
        :return: The files.
        """
        return await self._base_resource.get_children(File)

    async def get_file_object(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[bytes]:
        """Gets a file object.
        :param uuid: The UUID of the file to get.
        :param name: The alias of the file to get.
        :return: The file object.
        """
        file = await self.get_file(uuid, name)
        if not file:
            logger.warning(f"Could not get file {name}")
            return None

        await file.get()
        logger.info(f"Downloading file {str(file)}")
        return await self.api_client.request("GET", file["url"], to_tazawa=False)

    # Role methods
    get_role = get_role
    get_roles = get_roles
    create_role = create_role
    delete_role = delete_role



funnel_data = [
    {"value": 60, "name": "Third"},
    {"value": 40, "name": "Fourth"},
    {"value": 20, "name": "Fifth"},
    {"value": 80, "name": "Second"},
    {"value": 100, "name": "First"},
]



from shimoku import Client


class UsesClient:
    clients_dict = {}

    def assign_s(self, shimoku_client: Client):
        self.clients_dict["s"] = shimoku_client  # error


def action(shimoku_client: Client):
    uses_shimoku_client = UsesClient()
    uses_shimoku_client.assign_s(shimoku_client)



from typing import TYPE_CHECKING, Optional, Union
from pandas import DataFrame

from copy import deepcopy

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


def thin_variant(y: list[str], series_options: dict):
    len_y = len(y)
    dot_series_options = deepcopy(series_options)
    series_options["barWidth"] = 3
    series_options["barGap"] = "-100%"
    series_options["label"] = {"show": False}
    if len_y > 1:
        series_options["itemStyle"]["opacity"] = 0.75
    series_options = len_y * [series_options]
    dot_series_options["type"] = "scatter"
    dot_series_options["itemStyle"]["opacity"] = 1
    dot_series_options["itemStyle"]["borderWidth"] = 0
    dot_series_options["label"] = {"position": "top", "color": "inherit"}
    series_options.append(dot_series_options)
    y = y + y
    return y, series_options


def shadow_variant(series_options: dict):
    series_options["showBackground"] = True
    series_options["backgroundStyle"] = {"color": "rgba(180, 180, 180, 0.2)"}
    series_options["label"] = {"position": "top", "color": "inherit"}
    return series_options


def horizontal_clean_variant(common_options: dict):
    common_options["xAxis"][0].update(
        {
            "position": "top",
            "axisLabel": {
                "rotate": -45,
            },
        }
    )
    return common_options


def interpret_variant(variant: Optional[str]) -> tuple[Optional[str], Optional[str]]:
    if not variant:
        return None, None
    variant = variant.split(" ")
    if len(variant) == 1:
        if variant[0] in ["thin", "shadow"]:
            return None, variant[0]
        elif variant[0] in ["clean", "minimal"]:
            return variant[0], None
        raise ValueError(f"Invalid variant: {variant}")
    elif len(variant) == 2 and variant[0] in ["clean", "minimal"]:
        return variant[0], variant[1]
    else:
        raise ValueError(f"Invalid variant: {variant}")


async def bar_chart(
    self: "PlotLayer",
    data: Union[str, DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a bar chart"""
    series_options = get_common_series_options()
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    series_options["type"] = "bar"
    clean, variant = interpret_variant(variant)

    if variant == "thin":
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        y, series_options = thin_variant(y, series_options)
    elif variant == "shadow":
        series_options = shadow_variant(series_options)

    await self._create_trend_chart(
        echart_options=get_common_echart_options(),
        variant=clean,
        series_options=series_options,
        option_modifications=option_modifications,
        axes=x,
        values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
    )


async def stacked_bar_chart(
    self: "PlotLayer",
    data: Union[str, DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a stacked bar chart"""
    series_options = get_common_series_options()
    series_options.update(
        {
            "type": "bar",
            "areaStyle": {},
            "stack": "stack",
            "itemStyle": {"borderRadius": [0, 0, 0, 0]},
        }
    )
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    clean, variant = interpret_variant(variant)
    if variant == "thin":
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        len_y = len(y)
        y, series_options = thin_variant(y, series_options)
        series_options[-1]["stack"] = "dot_stack"
        series_options[-1]["symbolSize"] = 5
        series_options[-1]["label"]["position"] = "left"
        series_options = (
            series_options[:len_y]
            + [series_options[-1]] * (len_y - 1)
            + [deepcopy(series_options[-1])]
        )
        series_options[-1]["label"]["position"] = "top"
        del series_options[-1]["symbolSize"]
    else:
        len_y = len(y) if y else len(data_mapping_to_tuples.keys()) - 1
        if len_y > 1:
            series_options = [series_options] * (len_y - 1)
            series_options.append(deepcopy(series_options[-1]))
        else:
            series_options = [series_options]
        if variant == "shadow":
            series_options[-1] = shadow_variant(series_options[-1])
        series_options[-1]["itemStyle"]["borderRadius"] = [9, 9, 0, 0]

    await self._create_trend_chart(
        echart_options=get_common_echart_options(),
        variant=clean,
        axes=x,
        values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        series_options=series_options,
        option_modifications=option_modifications,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
    )


async def zero_centered_bar_chart(
    self: "PlotLayer",
    data: Union[str, DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a zero centered bar chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    series_options["type"] = "bar"
    common_options["xAxis"][0]["type"] = "value"
    series_options["itemStyle"]["borderRadius"] = [0, 0, 0, 0]
    del common_options["xAxis"][0]["data"]
    common_options["yAxis"][0].update(
        {
            "type": "category",
            "data": "#set_data#",
        }
    )
    clean, variant = interpret_variant(variant)
    if variant == "thin":
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        y, series_options = thin_variant(y, series_options)
    elif variant == "shadow":
        series_options = shadow_variant(series_options)

    await self._create_trend_chart(
        echart_options=common_options
        if not clean
        else horizontal_clean_variant(common_options),
        series_options=series_options,
        variant=clean,
        option_modifications=option_modifications,
        axes=x,
        values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
    )


async def horizontal_bar_chart(
    self: "PlotLayer",
    data: Union[str, DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a horizontal bar chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    series_options["type"] = "bar"
    common_options["xAxis"][0]["type"] = "value"
    series_options["itemStyle"]["borderRadius"] = [0, 9, 9, 0]
    del common_options["xAxis"][0]["data"]
    common_options["yAxis"][0].update(
        {
            "type": "category",
            "data": "#set_data#",
        }
    )
    clean, variant = interpret_variant(variant)
    if variant == "thin":
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        y, series_options = thin_variant(y, series_options)
        series_options[-1]["label"]["position"] = "right"
    elif variant == "shadow":
        series_options = shadow_variant(series_options)
        series_options["label"]["position"] = "right"

    await self._create_trend_chart(
        echart_options=common_options
        if not clean
        else horizontal_clean_variant(common_options),
        series_options=series_options,
        variant=clean,
        option_modifications=option_modifications,
        axes=x,
        values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
    )


async def stacked_horizontal_bar_chart(
    self: "PlotLayer",
    data: Union[str, DataFrame, list[dict]],
    order: int,
    x: str,
    y: Union[list[str], str, None] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a stacked horizontal bar chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options.update(
        {
            "type": "bar",
            "areaStyle": {},
            "stack": "stack",
            "itemStyle": {"borderRadius": [0, 0, 0, 0]},
        }
    )
    data_mapping_to_tuples = await self._choose_data(order, data=data)
    clean, variant = interpret_variant(variant)
    if variant == "thin":
        y = y if y else [k for k in data_mapping_to_tuples.keys() if k != x]
        len_y = len(y)
        y, series_options = thin_variant(y, series_options)
        series_options[-1]["stack"] = "dot_stack"
        series_options[-1]["symbolSize"] = 5
        series_options[-1]["label"]["position"] = "top"
        series_options = (
            series_options[:len_y]
            + [series_options[-1]] * (len_y - 1)
            + [deepcopy(series_options[-1])]
        )
        series_options[-1]["label"]["position"] = "right"
        del series_options[-1]["symbolSize"]
    else:
        len_y = len(y) if y else len(data_mapping_to_tuples.keys()) - 1
        series_options = [series_options] * (len_y - 1)
        series_options.append(deepcopy(series_options[-1]))
        series_options[-1]["itemStyle"]["borderRadius"] = [0, 9, 9, 0]
        if variant == "shadow":
            series_options[-1] = shadow_variant(series_options[-1])
            series_options[-1]["label"]["position"] = "right"

    common_options["xAxis"][0]["type"] = "value"
    del common_options["xAxis"][0]["data"]
    common_options["yAxis"][0].update(
        {
            "type": "category",
            "data": "#set_data#",
        }
    )

    await self._create_trend_chart(
        echart_options=common_options
        if not clean
        else horizontal_clean_variant(common_options),
        variant=clean,
        series_options=series_options,
        option_modifications=option_modifications,
        axes=x,
        values=y,
        data_mapping_to_tuples=data_mapping_to_tuples,
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
    )



import uuid
import ast
from typing import Dict, Optional
from shimoku.utils import IN_BROWSER
from shimoku.exceptions import ActionError
from shimoku.execution_logger import log_error
import warnings

import logging

logger = logging.getLogger(__name__)


class Scope:
    def __init__(self, parent_scope: Optional["Scope"] = None):
        self.lines_need_async = set()
        self.inner_scopes: Dict[str, Scope] = {}
        self.needs_async = False
        self.parent_scope: Optional[Scope] = parent_scope
        self.parameters: dict = {}

    def get_function(self, name: str):
        if name in self.inner_scopes:
            return self.inner_scopes[name]
        if self.parent_scope is not None:
            return self.parent_scope.get_function(name)
        return None

    def user_defined_scope(self):
        return (
            self.parent_scope is not None and self.parent_scope.parent_scope is not None
        )

    def enter_scope(self, name: str):
        if name not in self.inner_scopes:
            self.inner_scopes[name] = Scope(self)
        return self.inner_scopes[name]

    def exit_scope(self):
        return self.parent_scope


class ActionsChecker(ast.NodeVisitor):
    shimoku_client_attrs = [
        "universes",
        "workspaces",
        "boards",
        "menu_paths",
        "components",
        "data",
        "io",
        "activities",
        "plt",
        "set_workspace",
        "set_board",
        "set_menu_path",
        "run",
    ]

    SHIMOKU_CLIENT_NAME = "shimoku_client"

    def __init__(self):
        self.current_scope = Scope()
        self.in_class = False
        self.exists_action_function = False
        self.parameters = {}
        self.client_assigned = False
        self.shimoku_imported = False

    def is_shimoku_client(self, node):
        if isinstance(node, ast.Name) and node.id == self.SHIMOKU_CLIENT_NAME:
            return True
        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id == "Client"
        ):
            return True
        return False

    def visit_Import(self, node):
        if not self.current_scope.parent_scope:
            return
        for alias in node.names:
            if "asyncio" in alias.name:
                log_error(logger, "Cannot import asyncio module.", ActionError)

    def visit_ImportFrom(self, node):
        if node.module == "shimoku":
            if any(
                asname is not None for asname in [alias.asname for alias in node.names]
            ):
                log_error(
                    logger, "Cannot import shimoku modules with renaming.", ActionError
                )
            for alias in node.names:
                if alias.name == "Client":
                    self.shimoku_imported = True
        if not self.current_scope.parent_scope:
            return
        if node.module == "asyncio":
            log_error(logger, "Cannot import from asyncio module.", ActionError)

    def visit_Assign(self, node):
        for target in node.targets:
            if self.is_shimoku_client(node.value):
                if self.client_assigned:
                    log_error(logger, "Cannot assign Client instance", ActionError)
                self.client_assigned = True
            elif isinstance(target, ast.Name):
                if target.id == self.SHIMOKU_CLIENT_NAME:
                    log_error(
                        logger,
                        f"Cannot assign to {self.SHIMOKU_CLIENT_NAME} variable.",
                        ActionError,
                    )

        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        if node.name == "action":
            if self.in_class:
                log_error(
                    logger,
                    "Cannot define action function inside a class.",
                    ActionError,
                )
            if self.current_scope.user_defined_scope():
                log_error(
                    logger,
                    "Cannot define action function inside another function.",
                    ActionError,
                )
            if self.exists_action_function:
                log_error(
                    logger, "Cannot define more than one action function.", ActionError
                )
            if len(node.args.args) != 1:
                log_error(
                    logger,
                    "Action function must have exactly one parameter.",
                    ActionError,
                )
            if node.args.args[0].arg != self.SHIMOKU_CLIENT_NAME:
                log_error(
                    logger,
                    "Action function parameter must be named 'shimoku_client'.",
                    ActionError,
                )
            if (
                node.args.args[0].annotation
                and node.args.args[0].annotation.id != "Client"
            ):
                log_error(
                    logger,
                    "Action function parameter must be annotated with 'Client'.",
                    ActionError,
                )
            self.exists_action_function = True

        self.current_scope = self.current_scope.enter_scope(node.name)
        self.generic_visit(node)

        for arg in node.args.args:
            if self.in_class and arg.arg == "self":
                continue

            if not arg.annotation:
                log_error(
                    logger,
                    "Annotations are mandatory for function parameters.",
                    ActionError,
                )
            if isinstance(arg.annotation, ast.Subscript):
                if "Client" in ast.dump(arg.annotation):
                    log_error(
                        logger,
                        "Cannot use Client instance as a "
                        "part of a parameter type, use 'Client' instead.",
                        ActionError,
                    )
            elif arg.annotation.id == "Client":
                if arg.arg != self.SHIMOKU_CLIENT_NAME:
                    log_error(
                        logger,
                        f"Client parameter must be named '{self.SHIMOKU_CLIENT_NAME}' not '{arg.arg}'.",
                        ActionError,
                    )
            elif arg.arg == self.SHIMOKU_CLIENT_NAME:
                log_error(
                    logger,
                    f"{self.SHIMOKU_CLIENT_NAME} parameter must be annotated with "
                    f"'Client' not '{arg.annotation.id}'.",
                    ActionError,
                )
            self.current_scope.parameters[arg.arg] = arg.annotation.id

        self.current_scope = self.current_scope.exit_scope()

    def visit_AsyncFunctionDef(self, node):
        if not self.current_scope.parent_scope:
            self.current_scope = self.current_scope.enter_scope(node.name)
            self.generic_visit(node)
            self.current_scope = self.current_scope.exit_scope()
        else:
            log_error(logger, "Cannot use async functions.", ActionError)

    def visit_ClassDef(self, node):
        self.in_class = True
        self.generic_visit(node)
        self.in_class = False

    def visit_Return(self, node):
        if isinstance(node.value, ast.Name):
            if node.value.id == self.SHIMOKU_CLIENT_NAME:
                log_error(
                    logger,
                    f"Cannot return Client instance '{node.value.id}'.",
                    ActionError,
                )
        self.generic_visit(node)

    def visit_Call(self, node):
        self.generic_visit(node)

        # Check if the call is a method call that requires await
        if isinstance(node.func, ast.Attribute):
            aux_node = node.func
            while not isinstance(aux_node.value, ast.Name):
                if not isinstance(aux_node.value, ast.Attribute):
                    return
                aux_node = aux_node.value
            if (
                aux_node.value.id == self.SHIMOKU_CLIENT_NAME
                and aux_node.attr in self.shimoku_client_attrs
            ):
                # print(
                #     f"Client method call '{node.func.attr}' used at line {node.lineno} without await."
                # )
                self.current_scope.lines_need_async.add((node.lineno, node.col_offset))
                self.current_scope.needs_async = True

        if (
            not isinstance(node.func, ast.Name)
            or self.current_scope.get_function(node.func.id) is None
        ):
            return

        function_scope = self.current_scope.get_function(node.func.id)
        if function_scope.needs_async:
            self.current_scope.needs_async = True

        parameters = function_scope.parameters
        parameter_types = list(parameters.values())
        for i, arg in enumerate(node.args):
            if self.is_shimoku_client(arg) and not parameter_types[i] == "Client":
                log_error(
                    logger,
                    f"The value asigned to the parameter '{arg.id}' is the Client, "
                    f"but the function '{node.func.id}' expects a '{parameter_types[i]}' instance.",
                    ActionError,
                )
        for keyword in node.keywords:
            if (
                self.is_shimoku_client(keyword.value)
                and not parameters[keyword.arg] == "Client"
            ):
                log_error(
                    logger,
                    f"The value asigned to the parameter '{keyword.arg}' is the Client, "
                    f"but the function '{node.func.id}' expects a '{parameters[keyword.arg]}' instance.",
                    ActionError,
                )


class AwaitInserter(ast.NodeTransformer):
    def __init__(self, scope: Scope):
        self.current_scope = scope

    def visit_Call(self, node):
        self.generic_visit(node)
        if not self.current_scope.parent_scope:
            return node

        if (node.lineno, node.col_offset) in self.current_scope.lines_need_async:
            self.current_scope.needs_async = True
            return ast.copy_location(ast.Await(value=node), node)

        # Check if the call is a function call that requires await
        func_name = node.func.id if isinstance(node.func, ast.Name) else node.func.attr
        function_scope = self.current_scope.get_function(func_name)
        if function_scope is not None and function_scope.needs_async:
            self.current_scope.needs_async = True
            # print(
            #     f"Function call '{func_name}' used at line {node.lineno} without await."
            # )
            return ast.copy_location(ast.Await(value=node), node)

        return node

    def visit_FunctionDef(self, node):
        self.current_scope = self.current_scope.enter_scope(node.name)
        node = self.generic_visit(node)
        async_node = None
        if self.current_scope.needs_async:
            # Convert to an async function
            async_node = ast.AsyncFunctionDef(
                name=node.name,
                args=node.args,
                body=node.body,
                decorator_list=node.decorator_list,
                returns=node.returns,
                type_comment=node.type_comment,
            )
        self.current_scope = self.current_scope.exit_scope()
        if async_node is not None:
            return ast.copy_location(async_node, node)
        return node

    def visit_AsyncFunctionDef(self, node):
        self.current_scope = self.current_scope.enter_scope(node.name)
        node = self.generic_visit(node)
        self.current_scope = self.current_scope.exit_scope()
        return node


def print_code_with_line_numbers(code: str):
    code_lines = code.split("\n")
    for i, line in enumerate(code_lines):
        print(f"{i:3} | {line}")


def analyze_action_code(
    code: str,
    print_code: bool = False,
) -> ast.Module:
    code = code.replace("\r\n", "\n").replace("\\\n", "").replace("\\\r", "")
    code_lines = code.split("\n")
    main_uuid = str(uuid.uuid4()).replace("-", "_")

    script_code_lines = [
        "import asyncio",
        "from shimoku.actions_execution.front_connection import global_front_end_connection",
        "",
        "global_front_end_connection.js_snackbar = js_snackbar",
        "",
        f"async def main_{main_uuid}():",
        *[f"    {line}" for line in code_lines],
        "    shimoku_client = Client(",
        "        access_token=js_access_token,",
        "        universe_id=js_universe_id,",
        "        environment=js_environment,",
        "        async_execution=True,",
        "        verbosity='INFO',",
        "        retry_attempts=1",
        "    )",
        "    shimoku_client.set_workspace(js_workspace_id)",
        "    action(shimoku_client)",
        "",
    ]
    if IN_BROWSER:
        script_code_lines.append(
            f"asyncio.get_event_loop().create_task(main_{main_uuid}())",
        )
    else:
        script_code_lines.insert(
            -3, "    shimoku_client._async_pool.ACTIONS_TEST = True"
        )
        script_code_lines.append(
            f"asyncio.run(main_{main_uuid}())",
        )

    if print_code:
        for i, line in enumerate(script_code_lines):
            print(f"{i:3} | {line}")
    try:
        script_ast = ast.parse("\n".join(script_code_lines))
        analyzer = ActionsChecker()
        analyzer.visit(script_ast)
        if not analyzer.exists_action_function:
            log_error(logger, "Action function is not defined.", ActionError)
        if not analyzer.shimoku_imported:
            log_error(logger, "Shimoku's Client is not imported.", ActionError)
        modifier = AwaitInserter(analyzer.current_scope)
        modifier.visit(script_ast)
    except Exception as e:
        print_code_with_line_numbers("\n".join(script_code_lines))
        log_error(logger, f"Error while analyzing the code: {e}", ActionError)
        raise e
    return script_ast


def execute_action_code(
    code: str,
    print_code: bool = False,
    js_access_token: str = None,
    js_universe_id: str = "local",
    js_environment: str = "production",
    js_workspace_id: str = "local",
    js_snackbar: Optional[callable] = None,
):
    context = {
        "js_access_token": js_access_token,
        "js_universe_id": js_universe_id,
        "js_environment": js_environment,
        "js_workspace_id": js_workspace_id,
        "js_snackbar": js_snackbar,
    }
    script_ast = analyze_action_code(code, print_code)
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    exec(compile(script_ast, filename="shimoku_action", mode="exec"), context)
    warnings.resetwarnings()


def main():
    # with open('../../../tests/mockable_tests/test_plot_api.py', 'r') as f:
    with open("../../../tests/personal_test.py", "r") as f:
        code_lines = f.readlines()
    execute_action_code("\n".join(code_lines))


if __name__ == "__main__":
    main()



from shimoku import Client
from .data.default import data


def infographics(shimoku_client: Client):
    shimoku_client.set_menu_path("test-bentobox", "Infographics")
    title = "Lorem ipsum"
    text = (
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et "
        "dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip"
        "ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu"
        " fugiat nulla pariatur. "
    )

    shimoku_client.plt.infographics_text_bubble(
        order=0,
        cols_size=6,
        title=title,
        text=text,
        bubble_location="bottom",
        chart_parameters=dict(
            data=data,
            x="date",
            rows_size=16,
            option_modifications=dict(
                toolbox={"show": False},
                grid={
                    "left": "0%",
                    "right": "0%",
                    "top": "0%",
                    "bottom": "0%",
                    "containLabel": True,
                },
            ),
        ),
    )

    shimoku_client.plt.infographics_text_bubble(
        order=2,
        cols_size=6,
        title=title,
        text=text,
        chart_function=shimoku_client.plt.line,
        chart_parameters=dict(
            x="date",
            y="x",
            data=data,
            option_modifications=dict(
                toolbox={"show": False},
                grid={
                    "left": "0%",
                    "right": "0%",
                    "top": "0%",
                    "bottom": "0%",
                    "containLabel": True,
                },
            ),
        ),
    )

    shimoku_client.plt.infographics_text_bubble(
        order=4,
        cols_size=6,
        title=title,
        text=text,
        chart_function=shimoku_client.plt.shimoku_gauge,
        chart_parameters=dict(value=70, name="Gauge", rows_size=18, padding="0,0,0,0"),
    )
    stacked_data = [
        {
            "Weekday": "Mon",
            "Email": 120,
            "Union Ads": 132,
            "Video Ads": 101,
            "Search Engine": 134,
        },
        {
            "Weekday": "Tue",
            "Email": 220,
            "Union Ads": 182,
            "Video Ads": 191,
            "Search Engine": 234,
        },
        {
            "Weekday": "Wed",
            "Email": 150,
            "Union Ads": 232,
            "Video Ads": 201,
            "Search Engine": 154,
        },
        {
            "Weekday": "Thu",
            "Email": 820,
            "Union Ads": 932,
            "Video Ads": 901,
            "Search Engine": 934,
        },
        {
            "Weekday": "Fri",
            "Email": 120,
            "Union Ads": 132,
            "Video Ads": 101,
            "Search Engine": 134,
        },
        {
            "Weekday": "Sat",
            "Email": 220,
            "Union Ads": 182,
            "Video Ads": 191,
            "Search Engine": 234,
        },
        {
            "Weekday": "Sun",
            "Email": 150,
            "Union Ads": 232,
            "Video Ads": 201,
            "Search Engine": 154,
        },
    ]
    shimoku_client.plt.infographics_text_bubble(
        order=6,
        cols_size=6,
        rows_size=4,
        title=title,
        text=text,
        bubble_location="bottom",
        chart_function=shimoku_client.plt.stacked_bar,
        chart_parameters=dict(
            data=stacked_data,
            x="Weekday",
            x_axis_name="weekday",
            y_axis_name="visits",
            option_modifications=dict(
                toolbox={"show": False},
                grid={
                    "left": "0%",
                    "right": "0%",
                    "top": "0%",
                    "bottom": "0%",
                    "containLabel": True,
                },
            ),
        ),
        background_color="var(--color-stripe-light)",
    )

    shimoku_client.plt.infographics_text_bubble(
        order=8,
        title=title,
        text=text,
        bubble_location="right",
        chart_parameters=dict(
            data=data,
            x="date",
            option_modifications=dict(
                toolbox={"show": False},
                grid={
                    "left": "0%",
                    "right": "0%",
                    "top": "0%",
                    "bottom": "0%",
                    "containLabel": True,
                },
            ),
        ),
        background_url="https://images.unsplash.com/photo-1569982175971-d92b01cf8694?ixlib=rb-4.0.3&ixid="
        "MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=870&q=80",
    )

    form_groups = {
        "Personal information": [
            {
                "mapping": "name",
                "fieldName": "name",
                "inputType": "text",
            },
        ],
        "Other group": [
            {
                "mapping": "skills",
                "fieldName": "Skills",
                "options": ["Backend", "Frontend", "UX/UI", "Api Builder", "DevOps"],
                "inputType": "checkbox",
            },
        ],
    }

    shimoku_client.plt.infographics_text_bubble(
        order=10,
        cols_size=8,
        title=title,
        text=text,
        bubble_location="left",
        chart_function=shimoku_client.plt.generate_input_form_groups,
        chart_parameters=dict(
            form_groups=form_groups,
            dynamic_sequential_show=True,
            padding="5,1,0,0",
        ),
        background_color="var(--color-primary-light)",
    )

    shimoku_client.plt.infographics_text_bubble(
        order=12,
        cols_size=4,
        title=title,
        text=text,
        bubble_location="left",
        chart_function=shimoku_client.plt.shimoku_gauge,
        chart_parameters=dict(
            value=49,
            name="Gauge",
            padding="5,1,0,0",
            cols_size=10,
            color=3,
        ),
    )
    shimoku_client.plt.infographics_text_bubble(
        order=14,
        cols_size=6,
        title=title,
        text=text,
        bubble_location="left",
        image_url="default",
        image_size=60,
        background_color="var(--color-primary-light)",
        chart_parameters=dict(
            data=data,
            x="date",
            cols_size=10,
            option_modifications=dict(
                toolbox={"show": False},
                grid={
                    "left": "0%",
                    "right": "0%",
                    "top": "0%",
                    "bottom": "0%",
                    "containLabel": True,
                },
            ),
        ),
    )
    shimoku_client.plt.infographics_text_bubble(
        order=16,
        cols_size=6,
        rows_size=4,
        title=title,
        text=text,
        bubble_location="right",
        image_url="default",
        chart_function=shimoku_client.plt.table,
        chart_parameters=dict(
            data=data,
            rows_size=3,
            cols_size=10,
        ),
    )

    shimoku_client.plt.infographics_text_bubble(
        order=18,
        title=title,
        text=text,
        bubble_location="right",
        image_url="default",
        image_size=50,
        chart_function=shimoku_client.plt.table,
        chart_parameters=dict(
            data=stacked_data,
            cols_size=16,
            rows_size=3,
        ),
    )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
        AppCodeGen,
    )
    from shimoku.api.resources.report import Report


def code_gen_from_html_string(html_string: str):
    """Generate code for an html string.
    :param html_string: html string to generate code from
    :return: list of code lines
    """
    code_lines = []
    current_line = ""
    for c in html_string:
        if c in ["\n", "\r"]:
            code_lines.append(current_line)
            current_line = ""
        elif c == "<":
            code_lines.append(current_line)
            current_line = "<"
        elif c == ";" and len(current_line) > 60:
            code_lines.append(current_line + ";")
            current_line = ""
        else:
            current_line += c
    if current_line:
        code_lines.append(current_line)

    return [f'"{line}"' for line in code_lines if line]


async def code_gen_from_html(self: "AppCodeGen", report: "Report") -> list[str]:
    """Generate code for an html report.
    :param report: report to generate code from
    :return: list of code lines
    """

    html = report["chartData"][0]["value"].replace("'", "\\'").replace('"', '\\"')
    html_lines = ["    " + line for line in code_gen_from_html_string(html)]
    if not html_lines:
        return ["pass"]
    html_lines[-1] += ","
    code_lines = [
        "shimoku_client.plt.html(",
        f'    order={report["order"]},',
    ]
    if report["sizeColumns"] != 12:
        code_lines.append(f'    cols_size={report["sizeColumns"]},')
    if report["sizeRows"]:
        code_lines.append(f'    rows_size={report["sizeRows"]},')
    if report["sizePadding"] != "0,0,0,0":
        code_lines.append(f'    padding="{report["sizePadding"]}",')

    code_lines.extend([f"    html={html_lines[0][4:]}", *html_lines[1:], ")"])

    return code_lines



from shimoku import Client


def zero_centered_bar_chart(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "zero-centered-bar-test")

    shimoku_client.plt.zero_centered_bar(
        data="zero centered data",
        x="Name",
        y="y",
        order=0,
    )

    shimoku_client.plt.zero_centered_bar(
        data="zero centered data",
        x="Name",
        y=["y", "z", "a"],
        x_axis_name="Axis x",
        y_axis_name="Axis y",
        title="Title",
        order=1,
        rows_size=3,
        cols_size=10,
        padding="0,0,0,1",
    )



from typing import Optional, Dict, TYPE_CHECKING

from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.universe import Universe

import logging

logger = logging.getLogger(__name__)


class ActivityTemplate(Resource):
    _module_logger = logger
    resource_type = "activityTemplate"
    alias_field = ("name", "version")
    plural = "activityTemplates"

    def __init__(
        self,
        parent: "Universe",
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[Dict] = None,
    ):
        params = dict(
            name=alias if alias else "",
            version="",
            description="",
            availableModeCost={},
            inputSettings={},
            minRunInterval=0,
            enabled=False,
            provider="",
            tags=[],
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            params_to_serialize=["availableModeCost", "inputSettings"],
            params=params,
        )



from copy import copy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
        AppCodeGen,
    )
    from shimoku.api.resources.report import Report

from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_indicators import (
    code_gen_from_indicator,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_echarts import (
    code_gen_from_echarts,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_annotated_echart import (
    code_gen_from_annotated_echart,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_table import (
    code_gen_from_table,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_form import (
    code_gen_from_form,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_html import (
    code_gen_from_html,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_button import (
    code_gen_from_button,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_filter import (
    code_gen_from_filter,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_iframe import (
    code_gen_from_iframe,
)


def delete_default_properties(properties: dict, default_properties: dict) -> dict:
    """Delete default properties from a report.
    :param properties: properties of a report
    :param default_properties: default properties of a report
    :return: properties without default properties
    """
    properties = copy(properties)
    for key, value in default_properties.items():
        if key not in properties:
            continue
        if properties[key] == value:
            del properties[key]
        if isinstance(value, dict):
            if key not in properties:
                continue
            properties[key] = delete_default_properties(properties[key], value)
            if len(properties[key]) == 0:
                del properties[key]
    return properties


async def code_gen_from_other_reports(
    self: "AppCodeGen", report: "Report"
) -> list[str]:
    """Generate code for a report that is not a tabs group.
    :param report: report to generate code from
    :param is_last: whether the report is the last one
    :return: list of code lines"""
    code_lines = []

    properties = delete_default_properties(
        report["properties"], report.default_properties
    )
    del properties["hash"]

    if report["reportType"] == "INDICATOR":
        code_lines.extend(await code_gen_from_indicator(self, report, properties))
    elif report["reportType"] == "ECHARTS2":
        code_lines.extend(await code_gen_from_echarts(self, report, properties))
    elif report["reportType"] == "TABLE":
        code_lines.extend(await code_gen_from_table(self, report, properties))
    elif report["reportType"] == "FORM":
        code_lines.extend(await code_gen_from_form(self, report, properties))
    elif report["reportType"] == "HTML":
        code_lines.extend(await code_gen_from_html(self, report))
    elif report["reportType"] == "IFRAME":
        code_lines.extend(await code_gen_from_iframe(self, report))
    elif report["reportType"] == "ANNOTATED_ECHART":
        code_lines.extend(
            await code_gen_from_annotated_echart(self, report, properties)
        )
    elif report["reportType"] == "BUTTON":
        code_lines.extend(await code_gen_from_button(self, report))
    elif report["reportType"] == "FILTERDATASET":
        code_lines.extend(await code_gen_from_filter(self, report))
    else:
        code_lines.extend(
            [
                f"shimoku_client.add_report({report['reportType']}, order={report['order']}, data=dict())"
            ]
        )

    return code_lines



from typing import TYPE_CHECKING
from shimoku.code_gen.utils_code_gen import code_gen_from_dict

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
        AppCodeGen,
    )
    from shimoku.api.resources.report import Report


async def code_gen_from_indicator(
    self: "AppCodeGen", report: "Report", properties: dict
) -> list[str]:
    """Generate code for an indicator report.
    :param report: report to generate code from
    :param properties: properties of the report
    :return: list of code lines
    """
    report_dict = report.cascade_to_dict()
    if self._actual_bentobox:
        report_dict["sizeColumns"] += 1
    report_params = self.code_gen_report_params(report_dict)
    properties_code_lines = code_gen_from_dict(properties, 4)
    return [
        "shimoku_client.plt.indicator(",
        *report_params,
        f"    data={properties_code_lines[0][4:]}",
        *properties_code_lines[1:],
        ")",
    ]



from typing import Optional
import threading

from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter
from shimoku.cli.utils import choose_from_menu

from shimoku.actions_execution.execute_action import execute_action_code

from shimoku.ai.ai_layer import get_activity_template, ActivityTemplate, AILayer
import logging

logger = logging.getLogger(__name__)


def add_execute_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the get parser to a parser
    :param parser: Parser to add the get parser to
    :return: Get parser
    """
    params = {
        "name": "execute",
        "description": "Commands to execute activities",
    }
    if parser:
        execute_parser = parser.add_command(**params)
    else:
        execute_parser = CLIParser(**params)
    common_args = [
        CLIFuncParam(
            name="local-port",
            arg_type=int,
            arg_help="Local port to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="environment",
            arg_type=str,
            arg_help="Environment to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="access-token",
            arg_type=str,
            arg_help="Access token to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="universe-id",
            arg_type=str,
            arg_help="Universe ID to use",
            mandatory=False,
        ),
    ]

    module_functions = [action, ai_function]

    for func in module_functions:
        execute_parser.decor_add_command(common_args=common_args)(func)

    return execute_parser


async def ai_function(
    universe_api_key: Optional[str],
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    ai_function: str = CLIFuncParam(prompt=True),
    version: str = CLIFuncParam(default=None, mandatory=False),
    **kwargs,
):
    """Execute any ai function
    :param universe_api_key: The universe api key that the ai funcion will use
    :param workspace_id: The workspace id
    :param menu_path: The id or name of the menu path
    :param ai_function: The name of the ai function
    :param version: The version of the ai function
    """

    async def get_params(_ai_layer: AILayer, _activity_template: ActivityTemplate):
        """Get the parameters interactively
        :param _ai_layer: AILayer
        :param _activity_template: ActivityTemplate
        """
        params = {}
        input_settings = _activity_template["inputSettings"]
        for param_name, definition in input_settings.items():
            param_value = None
            while param_value is None:
                input_str = (
                    f"Enter value for {param_name}\n"
                    f"  type: {definition['datatype']}\n"
                    f"  description: {definition['description']})\n"
                )
                if not definition["mandatory"]:
                    input_str = input_str[:-1] + "Press enter to skip\n"
                if definition["datatype"] == "file":
                    input_file_names = [
                        file["file_name"]
                        for file in await _ai_layer.get_available_input_files()
                    ] + [""]
                    param_value = choose_from_menu(input_file_names, input_str)
                elif definition["datatype"] == "model":
                    model_names = [
                        model["file_name"]
                        for model in await _ai_layer.get_available_models()
                    ] + [""]
                    param_value = choose_from_menu(model_names, input_str)
                else:
                    param_value = input(input_str)
                    if param_value and definition["datatype"] != "str":
                        try:
                            param_value = eval(
                                f"{definition['datatype']}({param_value})"
                            )
                        except ValueError:
                            logger.error(f"Invalid value for {param_name}")
                            param_value = None

                if not param_value:
                    if definition["mandatory"]:
                        param_value = None
                        print(f"{param_name} is mandatory")
                    else:
                        continue

                params[param_name] = param_value

        return params

    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    universe = await resource_getter.get_universe()
    activity_template = await get_activity_template(universe, ai_function, version)
    ai_layer = await resource_getter.get_ai_layer()
    await ai_layer.generic_execute(
        ai_function=ai_function,
        version=version,
        universe_api_key=universe_api_key,
        **(await get_params(ai_layer, activity_template)),
    )


async def action(
    action: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Execute an action
    :param action: The id or name of the action
    """
    resource_getter = ResourceGetter(InitOptions(action=action, **kwargs))
    actions_layer = await resource_getter.get_actions_layer()
    action = await resource_getter.get_action()
    action_code = await actions_layer.get_action_code(action["id"])
    thread = threading.Thread(target=execute_action_code, args=(action_code,))
    thread.start()
    thread.join()


if __name__ == "__main__":
    import asyncio

    asyncio.run(add_execute_parser().parse_args())



import pandas as pd
from os import path
from shimoku import Client

data_path = path.join(path.dirname(path.abspath(__file__)), "data")


def scatter(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "scatter-test")

    shimoku_client.plt.scatter(
        data="main data",
        point_fields=[("x", "y")],
        order=0,
        cols_size=6,
        variant="clean",
    )
    scatter_data = pd.read_csv(path.join(data_path, "scatter_test.csv"))[
        [
            "AgeGroup1",
            "AgeGroup2",
            "AgeGroup3",
            "AllGroupAge",
            "WeightGroup1",
            "WeightGroup2",
            "WeightGroup3",
            "AllGroupWeight",
        ]
    ]
    shimoku_client.plt.scatter(
        data=scatter_data,
        point_fields=[
            ("AgeGroup1", "WeightGroup1"),
            ("AgeGroup2", "WeightGroup2"),
            ("AgeGroup3", "WeightGroup3"),
            ("AllGroupAge", "AllGroupWeight"),
        ],
        title="Age Weight correlation study",
        x_axis_name="Age",
        y_axis_name="Weight",
        order=1,
        rows_size=4,
        cols_size=6,
    )



tree_data = [
    {
        "name": "root",
        "value": 35,
        "children": [
            {
                "name": "Child A",
                "value": 9,
                "children": [
                    {"name": "Child A1", "value": 23},
                    {"name": "Child A2", "value": 72},
                    {"name": "Child A3", "value": 93},
                ],
            },
            {
                "name": "Child B",
                "value": 56,
                "children": [
                    {"name": "Child B1", "value": 39},
                    {"name": "Child B2", "value": 61},
                    {"name": "Child B3", "value": 71},
                ],
            },
            {
                "name": "Child C",
                "value": 100,
                "children": [
                    {"name": "Child C1", "value": 19},
                    {"name": "Child C2", "value": 66},
                    {"name": "Child C3", "value": 47},
                ],
            },
        ],
    }
]



from shimoku.cli import CLIParser, CLIFuncParam
from typing import Optional
from shimoku.exceptions import ActionError
from shimoku.actions_execution.execute_action import execute_action_code
import time
import subprocess
import threading
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
import ast

import logging

logger = logging.getLogger(__name__)


def add_listener_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the listener parser to a parser
    :param parser: Parser to add the listener parser to
    """

    params = {
        "name": "listen",
        "description": "Commands to execute continuously a file by listening to changes",
    }

    if parser:
        listener_parser = parser.add_command(**params)
    else:
        listener_parser = CLIParser(**params)

    common_args = [
        CLIFuncParam(
            name="path_to_code",
            arg_help="Path to the file to be listened to",
            arg_type=str,
            is_path=True,
            prompt=True,
        )
    ]

    module_functions = [script, action]

    for func in module_functions:
        listener_parser.decor_add_command(common_args=common_args)(func)

    return listener_parser


class FileChangeHandler(FileSystemEventHandler):
    def __init__(self, script_path, is_action=False):
        self.script_path = script_path
        self.is_action = is_action
        self.last_content = None

    def on_modified(self, event):
        if event.src_path == self.script_path:
            print(f"Detected change in file: {self.script_path}")
            self.process_file()

    def process_file(self):
        try:
            with open(self.script_path, "r") as file:
                current_content = file.read()

            if current_content == self.last_content:
                print("File content unchanged, ignoring event.")
                return

            self.last_content = current_content
            ast.parse(current_content)  # Check for syntax errors

            if not self.is_action:
                self.execute_script()
            else:
                self.execute_action()

        except SyntaxError as e:
            print("Could not execute file, syntax error:", e)
        except subprocess.CalledProcessError as e:
            print("Could not execute file:", e)
        except Exception as e:
            print(f"Error processing file: {e}")

    def execute_script(self):
        subprocess.run(f"python {self.script_path}", check=True, shell=True)
        print("Script executed successfully")

    def execute_action(self):
        thread = threading.Thread(target=execute_action_code, args=(self.last_content,))
        thread.start()
        thread.join()
        print("Action executed successfully")


def script(**kwargs):
    """
    Function to listen to changes in a script
    """
    path_to_code = kwargs["path_to_code"]
    event_handler = FileChangeHandler(path_to_code)
    observer = Observer()
    observer.schedule(event_handler, path=path_to_code, recursive=False)
    observer.start()
    print("Watching for file changes. Press Ctrl+C to stop.")
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def action(**kwargs):
    """
    Function to listen to changes in an action
    """
    path_to_code = kwargs["path_to_code"]
    event_handler = FileChangeHandler(path_to_code, is_action=True)
    observer = Observer()
    observer.schedule(event_handler, path=path_to_code, recursive=False)
    observer.start()
    print("Watching for file changes. Press Ctrl+C to stop.")
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



import datetime as dt
import pandas as pd
import asyncio
from copy import deepcopy
from typing import Optional, Union, TYPE_CHECKING, TypeVar
from numbers import Number


from ..base_resource import Resource
from shimoku.exceptions import DataError

if TYPE_CHECKING:
    from .app import App

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)

Mapping = TypeVar("Mapping", bound=Union[dict, list[str], str])


def get_column_types(data_point: dict, sort: Optional[dict]) -> dict:
    """Given a data point, return a dictionary with the column names as keys and their types as values
    :param data_point: a data point
    :param sort: a dictionary with the sort field and order
    """
    data_mapping = {}
    str_counter = 0
    float_counter = 0
    date_counter = 0
    for k, v in data_point.items():
        if str_counter > 50:
            raise ValueError("Too many string fields")
        if float_counter > 50:
            raise ValueError("Too many int fields")
        if date_counter > 50:
            raise ValueError("Too many date fields")
        if isinstance(v, str) or isinstance(v, bool):
            str_counter += 1
            data_mapping.update({k: f"stringField{str_counter}"})
        elif isinstance(v, Number):
            if sort and k == sort["field"]:
                data_mapping.update({k: "orderField1"})
                sort["field"] = "orderField1"
            else:
                float_counter += 1
                data_mapping.update({k: f"intField{float_counter}"})
        elif isinstance(v, (dt.date, dt.datetime, pd.Timestamp)):
            date_counter += 1
            data_mapping.update({k: f"dateField{date_counter}"})
            if sort and k == sort["field"]:
                sort["field"] = f"dateField{date_counter}"
        elif isinstance(v, dict):
            data_mapping.update({k: "customField1"})
        else:
            log_error(
                logger, f"Unknown value type {v} | Type {type(v)} | Key {k}", DataError
            )
    return data_mapping


def change_keys_df(df: pd.DataFrame, mapping: dict) -> list[dict]:
    """Given a data and a mapping, change the keys of the data to the values of the mapping
    :param df: data to change the keys
    :param mapping: mapping of the keys to change
    """
    df = df.rename(columns=mapping)
    for k in mapping.values():
        # test if all the values are of the same type
        if "int" in k or "order" in k:
            try:
                df[k] = df[k].astype(float)
            except ValueError:
                log_error(
                    logger,
                    f"Column {k} contains values that cannot be converted to float",
                    DataError,
                )
        elif "string" in k:
            try:
                df[k] = df[k].astype(str)
            except ValueError:
                log_error(
                    logger,
                    f"Column {k} contains values that cannot be converted to string",
                    DataError,
                )
        elif "date" in k:
            try:
                df[k] = pd.to_datetime(df[k]).dt.strftime("%Y-%m-%dT%H:%M:%SZ")
            except ValueError:
                log_error(
                    logger,
                    f"Column {k} contains values that cannot be converted to date ",
                    DataError,
                )

    return df.to_dict("records")


def convert_input_data_to_db_items(
    data: Union[pd.DataFrame, dict],
    sort: Optional[dict] = None,
    dump_whole: bool = False,
    column_types: Optional[dict] = None,
) -> Union[list[dict], dict]:
    """Given an input data, for all the keys of the data convert it to
     a Shimoku body parameter for data table

      stringField1: String
      stringField2: String
      stringField3: String
      stringField4: String
      intField1: Float
      intField2: Float
      intField3: Float
      intField4: Float
      dateField1: AWSDateTime
      customField1: AWSJSON

    Example
    ------------
    input
        data = [
         {'product': 'Matcha Latte', '2015': 43.3, '2016': 85.8, '2017': 93.7},
         {'product': 'Milk Tea', '2015': 83.1, '2016': 73.4, '2017': 55.1},
         {'product': 'Cheese Cocoa', '2015': 86.4, '2016': 65.2, '2017': 82.5},
         {'product': 'Walnut Brownie', '2015': 72.4, '2016': 53.9, '2017': 39.1}
        ]

    intermediate output
        d = {
            'product': 'stringField1',
            '2015': 'intField1',
            '2016': 'intField2',
            '2017': 'intField3'
        }

    output
        [
            {'stringField1': 'Matcha Latte',
             'intField1': 43.3,
             'intField2': 85.8,
             'intField3': 93.7
             },
            {'stringField1': 'Milk Tea',
             'intField1': 83.1,
             'intField2': 73.4,
             'intField3': 55.1
             },
            {'stringField1': 'Cheese Cocoa',
             'intField1': 86.4,
             'intField2': 65.2,
             'intField3': 82.5
             },
            {'stringField1': 'Walnut Brownie',
             'intField1': 72.4,
             'intField2': 53.9,
             'intField3': 39.1
             }
        ]
    """
    if dump_whole:
        return {"customField1": data}
    elif isinstance(data, pd.DataFrame):
        first_element = data.iloc[0].to_dict()
        column_mapping = (
            get_column_types(first_element, sort)
            if column_types is None
            else column_types
        )
        if data.isnull().values.any():
            log_error(
                logger,
                "Data contains null values please check for missing values",
                DataError,
            )
        return change_keys_df(data, column_mapping)
    else:
        log_error(logger, f"Unknown data type {type(data)}", DataError)


def convert_data_and_get_series_name(
    data: pd.DataFrame, field: str
) -> tuple[pd.DataFrame, str]:
    """Convert data to a format that can be used by the API and get the series name of a field.
    :param data: data to convert
    :param field: field to get the series name
    :return: converted data and the converted series name
    """
    converted_data = pd.DataFrame(convert_input_data_to_db_items(data))

    converted_data_columns = converted_data.columns.to_list()
    data_columns = data.columns.to_list()
    series_name = converted_data_columns[data_columns.index(field)]

    return converted_data, series_name


class DataSet(Resource):
    """
    DataSet resource class
    """

    _module_logger = logger
    resource_type = "dataSet"
    alias_field = "name"
    plural = "dataSets"

    class DataPoint(Resource):
        _module_logger = logger
        resource_type = "data"
        plural = "datas"

        def __init__(
            self,
            parent: "DataSet",
            uuid: Optional[str] = None,
            db_resource: Optional[dict] = None,
        ):
            params = dict(
                dataSetId=parent["id"],
                orderField1=None,
                description=None,
                customField1=None,
            )
            for i in range(1, 6):
                params.update({f"dateField{i}": None})
            for i in range(1, 51):
                params.update({f"stringField{i}": None})
                params.update({f"intField{i}": None})

            super().__init__(
                parent=parent,
                db_resource=db_resource,
                uuid=uuid,
                params=params,
                params_to_serialize=["customField1"],
            )

        async def delete(self):
            """Delete the data point"""
            return await self._base_resource.delete()

        async def update(self):
            """Update the data point"""
            return await self._base_resource.update()

    def __init__(
        self,
        parent: "App",
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = dict(
            name=alias,
            columns=None,
        )
        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            children=[self.DataPoint],
            check_params_before_creation=["name"],
            params=params,
            params_to_serialize=["columns"],
        )

    def delete(self):
        """Delete the data point"""
        return self._base_resource.delete()

    async def update(self):
        """Update the data point"""
        return await self._base_resource.update()

    # DataPoint methods
    async def create_data_points(
        self,
        data_points: Union[pd.DataFrame, dict],
        sort: Optional[dict] = None,
        dump_whole: bool = False,
        column_types: Optional[dict] = None,
    ) -> tuple[Mapping, dict]:
        """Create data points for a report
        :param data_points: data points to be created
        :param sort: sort parameter
        :param dump_whole: whether to dump the whole data into a single field
        :param column_types: precomputed column types
        :return: mapping
        """
        copy_sort = deepcopy(sort) if sort else None
        converted_data_points = convert_input_data_to_db_items(
            data_points, copy_sort, dump_whole=dump_whole, column_types=column_types
        )
        if isinstance(converted_data_points, dict):
            converted_data_points = [converted_data_points]
        await self._base_resource.create_children_batch(
            self.DataPoint, converted_data_points, unit=" data points"
        )
        keys = [
            k
            for k in converted_data_points[0].keys()
            if not copy_sort or copy_sort["field"] != k
        ]
        if (
            self.api_client.playground
            and isinstance(data_points, (pd.DataFrame, list))
            and len(data_points) > 0
        ):
            if isinstance(data_points, pd.DataFrame):
                data_points = data_points.to_dict("records")
            self["columns"] = {
                k: m
                for k, m in zip(data_points[0].keys(), converted_data_points[0].keys())
            }
            await self.update()
        return keys, copy_sort

    async def delete_data_points(self):
        """Delete data points"""
        self.clear()
        data_points = await self._base_resource.get_children(self.DataPoint)
        if data_points:
            await asyncio.gather(
                *[
                    self._base_resource.delete_child(self.DataPoint, uuid=dp["id"])
                    for dp in data_points
                ]
            )

    async def delete_data_point(self, uuid: str):
        """Delete a data point"""
        self.clear()
        await self._base_resource.delete_child(self.DataPoint, uuid=uuid)

    async def _update_data_point(self, uuid: str, data: dict):
        """Update a data point"""
        data_point = await self._base_resource.get_child(self.DataPoint, uuid=uuid)
        data_point_as_dict = data_point.cascade_to_dict()
        data_point.set_params(**{
            k: data[k] for k in data_point_as_dict
            if k in data and
            data[k] != data_point_as_dict[k]
        })
        await data_point.update()

    async def update_data_point(self, uuid: str, data: dict):
        """Update a data point"""
        self.clear()
        await self._update_data_point(uuid, data)

    async def linear_update_data_points(self, data_points: list[dict]):
        """Update data points linearly, this function only is more efficient when
        the data points to update can be considered constant, otherwise deleting and creating
        can be more efficient

        data has to follow the schema of the dataset! and it has to have an orderField1

        :param data_points: data points to update
        """
        self.clear()

        if 'orderField1' not in data_points[0]:
            log_error(logger, "The data points must have an orderField1", DataError)

        current_data_points = await self._base_resource.get_children(self.DataPoint)
        current_data_points = [dp.cascade_to_dict() for dp in current_data_points]

        try:
            current_data_points = sorted(current_data_points, key=lambda x: x["orderField1"])
        except TypeError:
            log_error(
                logger,
                "The orderField1 has to be sortable and all the values have to be the same type",
                DataError
            )
        tasks = []

        if len(data_points) < len(current_data_points):
            current_data_points, current_data_points_to_delete = (
                current_data_points[:len(data_points)],
                current_data_points[len(data_points):]
            )
            for del_dp in current_data_points_to_delete:
                tasks.append(self.delete_data_point(del_dp["id"]))

        elif len(data_points) > len(current_data_points):
            data_points, data_points_to_create = (
                data_points[:len(current_data_points)],
                data_points[len(current_data_points):]
            )
            tasks.append(
                self.create_data_points(
                    data_points=pd.DataFrame(data_points_to_create),
                    sort={"field": "orderField1"}
                )
            )

        for updated_dp, dp in zip(data_points, current_data_points):
            if any([updated_dp[k] != dp[k] for k in updated_dp]):
                tasks.append(self._update_data_point(dp["id"], updated_dp))

        await asyncio.gather(*tasks)

    async def get_data_points(self, limit: Optional[int] = None):
        """Get data points"""
        self.clear()
        return await self._base_resource.get_children(self.DataPoint, limit=limit)

    async def get_one_data_point(self) -> Optional["DataSet.DataPoint"]:
        """Get one data point"""
        self.clear()
        data_points = await self._base_resource.get_children(self.DataPoint, limit=1)
        return data_points[0] if data_points else None



from shimoku.utils import create_function_name
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_other import (
    delete_default_properties,
)
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_tabs import (
    code_gen_tabs_and_other,
)
from typing import TYPE_CHECKING
from shimoku.code_gen.utils_code_gen import code_gen_from_value
from shimoku.api.resources.reports.modal import Modal

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
        AppCodeGen,
    )


async def code_gen_from_modal(self: "AppCodeGen", tree: dict) -> list[str]:
    """Generate code for a modal.
    :param tree: tree of reports
    :return: list of code lines
    """
    # code_lines = (await code_gen_tabs_functions(tree['tab_groups']))
    code_lines = []
    modal: Modal = tree["modal"]
    properties = delete_default_properties(
        modal["properties"], Modal.default_properties
    )
    properties["modal_name"] = tree["name"]
    del properties["hash"]
    if "reportIds" in properties:
        del properties["reportIds"]
    if "open" in properties:
        if properties["open"]:
            properties["open_by_default"] = True
        del properties["open"]
    code_lines.extend(
        [
            "shimoku_client.plt.set_modal(",
            *[f"    {k}={code_gen_from_value(v)}," for k, v in properties.items()],
            ")",
        ]
    )
    code_lines.extend(await code_gen_tabs_and_other(self, tree))
    code_lines.extend(["", "shimoku_client.plt.pop_out_of_modal()"])
    return code_lines


async def code_gen_modals_functions(self: "AppCodeGen", path: str) -> list[str]:
    """Generate code for modals.
    :param path: path to the modals
    :return: list of code lines
    """
    code_lines = []
    for modal in self._code_gen_tree.tree[path]["modals"]:
        modal_code_lines = await code_gen_from_modal(self, modal)
        code_lines.extend(
            [
                "",
                f'def modal_{create_function_name(modal["name"])}'
                f"(shimoku_client: Client):",
                *["    " + line for line in modal_code_lines],
            ]
        )
    return code_lines



from typing import Optional, TYPE_CHECKING

import pandas as pd

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def speed_gauge_chart(
    self: "PlotLayer",
    order: int,
    name: str,
    value: int,
    min_value: int,
    max_value: int,
    title: Optional[str] = None,
    rows_size: Optional[int] = 2,
    cols_size: Optional[int] = 3,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a speed gauge chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["yAxis"]
    del common_options["xAxis"]
    del common_options["grid"]
    del common_options["toolbox"]["feature"]["magicType"]
    common_options["grid"] = {"top": "1%", "left": "1%", "right": "1%"}
    common_options["tooltip"]["trigger"] = "item"
    series_options.update(
        {
            "type": "gauge",
            "min": min_value,
            "max": max_value,
            "progress": {"show": True, "width": 18},
            "axisLine": {"lineStyle": {"width": 18}},
            "axisTick": {"show": False},
            "splitLine": {"length": 15, "lineStyle": {"width": 2, "color": "#999"}},
            "axisLabel": {"distance": 25, "color": "#999", "fontSize": 10},
            "anchor": {
                "show": True,
                "showAbove": True,
                "size": 25,
                "itemStyle": {"borderWidth": 10},
            },
            "title": {"show": True, "offsetCenter": [0, "45%"], "fontSize": 20},
            "detail": {
                "valueAnimation": True,
                "fontSize": 20,
                "offsetCenter": [0, "70%"],
            },
        }
    )
    common_options["series"] = [series_options]

    await self._create_echart(
        fields=[{"name": "name", "value": "value"}],
        options=common_options,
        option_modifications=option_modifications,
        data_mapping_to_tuples=await self._choose_data(
            order, pd.DataFrame([{"name": name, "value": value}])
        ),
        order=order,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )



from shimoku import Client
from copy import deepcopy
from .data.radar import radar_data


def radar(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "radar-test")

    shimoku_client.plt.radar(
        data=radar_data, names="name", order=1, rows_size=2, cols_size=6
    )

    aux_radar_data = deepcopy(radar_data)
    aux_radar_data[0]["max"] = 90
    aux_radar_data[1]["max"] = 70
    aux_radar_data[2]["max"] = 70
    aux_radar_data[3]["max"] = 80

    shimoku_client.plt.radar(
        data=aux_radar_data,
        names="name",
        order=2,
        rows_size=2,
        cols_size=12,
        max_field="max",
        fill_area=True,
    )



from shimoku import Client
from .data.summary_line import summary_data


def summary_line(shimoku_client: Client):
    shimoku_client.set_menu_path("test-bentobox", "summary-line")

    shimoku_client.plt.line_with_summary(
        data=summary_data,
        order=4,
        x="date",
        y="x",
        title="Total",
        description="Today",
        value=7.94,
    )



from typing import Optional, Union, TYPE_CHECKING

from shimoku.exceptions import BentoboxError
from shimoku.plt.utils import interpret_color, ShimokuPalette

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


def gauge_indicator(
    self: "PlotLayer",
    order: int,
    value: int,
    title: Optional[str] = "",
    description: Optional[str] = "",
    cols_size: Optional[int] = 6,
    rows_size: Optional[int] = 1,
    color: Union[str, int] = 1,
):
    """Create a gauge indicator"""
    bentobox_data = {
        "bentoboxId": f"_{order}",
        "bentoboxOrder": order,
        "bentoboxSizeColumns": cols_size,
        "bentoboxSizeRows": rows_size,
    }

    if self._bentobox_data:
        log_error(
            logger,
            "The gauges group uses a bentobox, so it cannot be used inside another bentobox. Please"
            " pop out of the current bentobox before using this function.",
            BentoboxError,
        )
    self._bentobox_data = bentobox_data

    indicator_data = [
        {
            "description": description,
            "title": title,
            "value": "",
            "color": "",
            "align": "left",
        }
    ]

    self.indicator(
        data=indicator_data,
        order=order,
        rows_size=8,
        cols_size=15,
        padding="1, 1, 0, 1",
    )

    color = interpret_color(color)

    options = {
        "grid": {
            "left": ["5%"],
            "right": ["5%"],
            "top": ["5%"],
            "bottom": ["5%"],
            "containLabel": True,
        },
        "series": [
            {
                "data": "#set_data#",
                "type": "gauge",
                "startAngle": 360,
                "endAngle": 0,
                "radius": "80%",
                "center": ["50%", "40%"],
                "min": 0,
                "max": 100,
                "pointer": {
                    "show": False,
                },
                "progress": {
                    "show": True,
                    "width": 20,
                    "overlap": False,
                    "roundCap": False,
                    "clip": False,
                    "itemStyle": {
                        "borderWidth": 0,
                        "color": color,
                    },
                },
                "splitLine": {
                    "show": False,
                },
                "axisLine": {"lineStyle": {"width": 20}},
                "axisTick": {"show": False},
                "axisLabel": {
                    "show": False,
                },
                "title": {
                    "show": False,
                    "fontSize": 16,
                    "fontFamily": "Rubik",
                },
                "detail": {
                    "fontSize": 24,
                    "fontFamily": "Rubik",
                    "font": "inherit",
                    "color": ShimokuPalette.BLACK.value,
                    "borderColor": "auto",
                    "borderWidth": 0,
                    "formatter": "{value}%",
                    "valueAnimation": True,
                    "offsetCenter": ["0", "60"],
                },
            }
        ],
    }

    self.free_echarts(
        data=[{"value": value}],
        fields=[{"value": "value"}],
        options=options,
        order=order + 1,
        rows_size=6,
        cols_size=7,
        padding="1, 0, 1, 0",
    )

    self.pop_out_of_bentobox()



html_data = (
    "<p style='background-color: #daf4f0';>"
    "Comparing the results of predictions that happened previous "
    "periods vs reality, so that you can measure the accuracy of our predictor"
    "</p>"
)



""""""
import pandas as pd
from pandas import DataFrame

from typing import Optional, Union

from shimoku.api.utils import validate_data_is_pandarable
from shimoku.api.resources.app import App
from shimoku.api.resources.data_set import convert_input_data_to_db_items
from shimoku.exceptions import DataError

import logging
from shimoku.execution_logger import log_error, ClassWithLogging

logger = logging.getLogger(__name__)


class DataSetsLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the data set level.
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(self, app: Optional[App]):
        self._app = app

    async def append_to_data_set(
        self,
        data: Union[list[dict], DataFrame],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> str:
        """
        Appends data to a dataset, if the dataset does not exist it will be created.
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param data: data to be stored in the dataset
        :return: dataset id
        """

        df = validate_data_is_pandarable(data)
        df["sort_values"] = range(len(df))

        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        data_point = await data_set.get_one_data_point()

        if data_point:
            converted_data = convert_input_data_to_db_items(df, sort={"field": "sort_values"})
            data_point_keys = [col for col in data_point if data_point[col] is not None]
            db_items_keys = set(converted_data[0].keys())
            if [key for key in db_items_keys if key not in data_point_keys]:
                log_error(
                    logger,
                    "The input data has different fields than the existing data set.",
                    DataError,
                )

        await self._app.append_data_to_data_set(
            df, uuid=uuid, name=name, sort={"field": "sort_values"}
        )

        return data_set["id"]

    async def delete_data_set(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Delete a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        await self._app.delete_data_set(uuid=uuid, name=name)

    async def delete_data_from_data_set(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Delete all data from a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        await data_set.delete_data_points()

    async def replace_data_from_data_set(
        self,
        data: Union[list[dict], DataFrame],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Replace the data in a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param data: data to be stored in the dataset
        """
        df = validate_data_is_pandarable(data)
        df["sort_values"] = range(len(df))

        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        await data_set.delete_data_points()

        await self._app.append_data_to_data_set(
            df, uuid=uuid, name=name, sort={"field": "sort_values"}
        )

    async def delete_data_point_from_data_set(
        self,
        data_point_id: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Delete a data point from a dataset in the menu path
        :param data_point_id: id of the data point
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        await data_set.delete_data_point(data_point_id)

    @staticmethod
    async def get_columns_mapping_from_data_point(data: dict) -> dict:
        """Get the columns mapping from a data point in the menu path,
        assuming that the null values are empty strings.

        :param data: data to be stored in the dataset
        :return: columns mapping
        """
        data = {k: v if v is not None else ""
                for k, v in data.items()}
        converted_data_point = convert_input_data_to_db_items(pd.DataFrame([data]))[0]
        return {column: mapping for column, mapping in zip(data.keys(), converted_data_point.keys())}

    async def update_data_point_from_data_set(
        self,
        data_point_id: str,
        data: dict,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Update a data point from a dataset in the menu path,
        data has to follow the schema of the dataset!

        e.g. if your dataset has columns 'id', 'name' and 'age',
        the format to update a data point should be:
        {'stringField2': 'new_name', 'intField1': 25}
             ^ because 'id' is the first string column and you wouldn't want to update it

        suggestion: use get_columns_mapping_from_data_point
        to get the columns mapping from a data point

        :param data_point_id: id of the data point
        :param data: data to be stored in the dataset
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        await data_set.update_data_point(
            data_point_id,
            data={k: v for k, v in data.items() if v is not None}
        )

    async def linear_update_data_from_data_set(
        self,
        data: Union[list[dict], DataFrame],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Update the data in the dataset following a linear update strategy,
        creates the new data points if they do not exist, deletes the old ones if they do not exist in the new data,
        and linearly checks if the data points have changed and updates them.
        This makes this function more efficient for small changes than replace_data_from_data_set.
        But if the changes are linear in proportion to the size of the dataset, this function can be
        slower than replace_data_from_data_set.

        :param data: data to be stored in the dataset
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        df = validate_data_is_pandarable(data)
        df["sort_values"] = range(len(df))
        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        data_points = convert_input_data_to_db_items(df, sort={"field": "sort_values"})
        await data_set.linear_update_data_points(data_points)

    async def get_data_from_data_set(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> list[dict]:
        """Get the data in a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param limit: limit of the data points to be returned
        :return: data in the dataset
        """
        data_set = await self._app.get_data_set(
            uuid=uuid, name=name, create_if_not_exists=False
        )
        if not data_set:
            logger.warning(f"Data set {name if name else uuid} does not exist.")
            return []
        return [
            {k: v for k, v in dp.cascade_to_dict().items() if v is not None}
            for dp in await data_set.get_data_points(limit)
        ]

    async def get_data_set_metadata(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Get the metadata of a dataset in the menu path
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :return: metadata of the dataset
        """
        data_set = await self._app.get_data_set(uuid=uuid, name=name)
        return data_set.cascade_to_dict()



from shimoku.actions_execution.execute_action import execute_action_code
from shimoku.exceptions import ActionError, CacheError
from utils import initiate_shimoku
from os import path, listdir
import unittest

correct_scripts_path = path.join(path.dirname(__file__), "correct_action_scripts")
error_scripts_path = path.join(path.dirname(__file__), "error_action_scripts")


class TestActions(unittest.TestCase):
    def test_action_execute(self):
        for correct_action_script in listdir(correct_scripts_path):
            execute_action_code(
                open(path.join(correct_scripts_path, correct_action_script)).read(),
                True,
            )

        for code_with_errors in listdir(error_scripts_path):
            print()
            with self.assertRaises(ActionError):
                execute_action_code(code_with_errors, True)

    def test_create_action(self):
        correct_action_script = path.join(correct_scripts_path, "correct_action.py")
        code = open(correct_action_script).read()
        shimoku_client = initiate_shimoku()
        shimoku_client.actions.create_action(
            "test action",
            "test action description",
            code=code,
            overwrite=True,
            libraries=["pandas==1.5.3"],
        )
        with self.assertRaises(CacheError):
            shimoku_client.actions.create_action(
                "test action",
                "test action description",
                code=code,
                overwrite=False,
                libraries=["pandas==1.5.3"],
            )
        action_metadata = shimoku_client.actions.get_action_metadata(name="test action")

        assert shimoku_client.actions.get_action_code(name="test action") == code
        assert action_metadata["description"] == "test action description"
        assert action_metadata["name"] == "test action"
        assert action_metadata["pythonLibraries"] == ["pandas==1.5.3"]

        shimoku_client.actions.delete_action(name="test action")

        with self.assertRaises(ActionError):
            shimoku_client.actions.get_action_metadata(name="test action")



from ..report import Report

import asyncio
import logging

logger = logging.getLogger(__name__)


class Modal(Report):
    """Modal report class"""

    report_type = "MODAL"

    default_properties = dict(
        **Report.default_properties,
        open=False,
        reportIds=[],
        width=60,
        height=50,
    )

    def __init__(self, *args, **kwargs):
        self.dirty = False
        super().__init__(*args, **kwargs)

    async def update(self, *args, **kwargs) -> bool:
        """Update the modal on the server
        :return: True if the modal was updated, False otherwise
        """
        if not self.dirty:
            return False
        self.dirty = False
        await super().update()
        return True

    async def delete(self):
        """Delete the modal from the server and all its children"""
        await asyncio.gather(
            *[
                self._base_resource.parent.delete_report(rd_id)
                for rd_id in self["properties"]["reportIds"]
            ]
        )
        await super().delete()

    def add_report(self, report: Report):
        """Add report to the modal without saving it to the server
        :param report: report to add
        """
        if report["id"] in self["properties"]["reportIds"]:
            logger.warning(f"Report {report['id']} already in modal")
            return

        self["properties"]["reportIds"].append(report["id"])
        self.dirty = True

    def remove_report(self, report: Report):
        """Remove report from the modal without saving it to the server
        :param report: report to remove
        """

        if report["id"] not in self["properties"]["reportIds"]:
            logger.warning(f"Report {report['id']} not in modal")
            return

        self["properties"]["reportIds"].remove(report["id"])
        self.dirty = True

    def has_report(self, report: Report):
        """Check if report is in the modal
        :param report: report to check
        """
        return report["id"] in self["properties"]["reportIds"]

    def clear_content(self):
        """Remove all reports from the tabs group without saving it to the server"""
        self["properties"]["reportIds"] = []
        self.dirty = True



from shimoku.api.resources.reports.tabs_group import TabsGroup
from shimoku.utils import create_function_name

from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.report_types_code_gen.code_gen_from_other import (
    code_gen_from_other_reports,
    delete_default_properties,
)
from shimoku.code_gen.utils_code_gen import code_gen_from_value
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.code_gen_from_reports import (
        AppCodeGen,
    )


async def code_gen_tabs_and_other(self: "AppCodeGen", tree: dict) -> list[str]:
    """Generate code for tabs and other components.
    :param tree: tree of reports
    :return: list of code lines
    """
    code_lines: list[str] = []
    reports_ordered = sorted(
        tree["other"] + tree["tab_groups"], key=lambda x: x["order"]
    )
    for i, report in enumerate(reports_ordered):
        report_obj = report if not isinstance(report, dict) else report["tabs_group"]
        if len(report_obj["bentobox"]):
            if (
                self._actual_bentobox is None
                or self._actual_bentobox["bentoboxId"]
                != report_obj["bentobox"]["bentoboxId"]
            ):
                self._actual_bentobox = report_obj["bentobox"]

                cols_size = self._actual_bentobox["bentoboxSizeColumns"]
                rows_size = self._actual_bentobox["bentoboxSizeRows"]
                code_lines.extend(
                    [
                        "",
                        f"shimoku_client.plt.set_bentobox(cols_size={cols_size}, rows_size={rows_size})",
                    ]
                )
        elif self._actual_bentobox is not None:
            self._actual_bentobox = None
            code_lines.append("shimoku_client.plt.pop_out_of_bentobox()")

        if isinstance(report, dict):
            code_lines.extend(
                [
                    "",
                    f'tabs_group_{create_function_name(report["name"])}(shimoku_client)',
                ]
            )
        else:
            code_lines.extend(await code_gen_from_other_reports(self, report))

        if i == len(reports_ordered) - 1 and self._actual_bentobox is not None:
            self._actual_bentobox = None
            code_lines.append("shimoku_client.plt.pop_out_of_bentobox()")

    return code_lines


async def code_gen_from_tabs_group(
    self: "AppCodeGen", tree: dict, is_last: bool = False
) -> list[str]:
    """Generate code for a tabs group.
    :param tree: tree of reports
    :param is_last: whether the tabs group is the last one
    :return: list of code lines
    """
    code_lines = []
    tabs_keys = list(tree["tabs"].keys())
    if len(tabs_keys) == 0:
        return ["pass"]
    tabs_group: TabsGroup = tree["tabs_group"]
    tabs_index = (tree["name"], tabs_keys[0])
    parent_tabs_index = tree["parent_tabs_index"]
    properties = delete_default_properties(
        tabs_group["properties"], TabsGroup.default_properties
    )
    del properties["hash"]
    if "tabs" in properties:
        del properties["tabs"]
    if "variant" in properties:
        properties["just_labels"] = True
        del properties["variant"]

    for tab in tree["tabs"]:
        code_lines.extend(
            [
                "",
                f"def tab_{create_function_name(tabs_index[0])}_{create_function_name(tab)}():",
            ]
        )
        # tab_code = await code_gen_tabs_functions(tree['tabs'][tab]['tab_groups'])
        tab_code = await code_gen_tabs_and_other(self, tree["tabs"][tab])
        if not tab_code:
            tab_code = ["pass"]
        code_lines.extend([f"    {line}" for line in tab_code])

    code_lines.extend(
        [
            "",
            "shimoku_client.plt.set_tabs_index(",
            f'    tabs_index=("{tabs_index[0]}", "{tabs_index[1]}"), order={tabs_group["order"]}, ',
        ]
    )
    if tabs_group["sizeColumns"]:
        code_lines.append(f'    cols_size={tabs_group["sizeColumns"]},')
    if tabs_group["sizeRows"]:
        code_lines.append(f'    rows_size={tabs_group["sizeRows"]},')
    if tabs_group["sizePadding"]:
        code_lines.append(f'    padding="{tabs_group["sizePadding"]}",')
    if parent_tabs_index:
        code_lines.extend([f"    parent_tabs_index={parent_tabs_index},"])
    code_lines.extend(
        [f"    {k}={code_gen_from_value(v)}," for k, v in properties.items()]
    )
    code_lines.extend([")"])

    for tab in tree["tabs"]:
        code_lines.extend(["", f'shimoku_client.plt.change_current_tab("{tab}")'])
        code_lines.extend(
            [f"tab_{create_function_name(tabs_index[0])}_{create_function_name(tab)}()"]
        )

    if parent_tabs_index:
        if not is_last:
            code_lines.extend(
                [
                    "",
                    f'shimoku_client.plt.set_tabs_index(("{parent_tabs_index[0]}", "{parent_tabs_index[1]}"))',
                ]
            )
    else:
        code_lines.extend(["", "shimoku_client.plt.pop_out_of_tabs_group()"])

    return code_lines


async def code_gen_tabs_functions(self, path: str) -> list[str]:
    """Generate code for tabs groups functions.
    :return: list of code lines
    """
    code_lines = []
    if path not in self._code_gen_tree.all_tab_groups:
        return code_lines
    for tabs_group in self._code_gen_tree.all_tab_groups[path]:
        tab_code_lines = await code_gen_from_tabs_group(self, tabs_group)
        code_lines.extend(
            [
                "",
                f'def tabs_group_{create_function_name(tabs_group["name"])}'
                f"(shimoku_client: Client):",
                *["    " + line for line in tab_code_lines],
            ]
        )
    return code_lines



from typing import Optional
from shimoku.actions_execution.execute_action import print_code_with_line_numbers
from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter
from shimoku.cli.utils import display_dict, save_as_file
import logging

logger = logging.getLogger(__name__)


def add_get_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the get parser to a parser
    :param parser: Parser to add the get parser to
    :return: Get parser
    """
    params = {"name": "get", "description": "Commands to get resources"}
    if parser:
        get_parser = parser.add_command(**params)
    else:
        get_parser = CLIParser(**params)

    common_args = [
        CLIFuncParam(
            name="local-port",
            arg_type=int,
            arg_help="Local port to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="environment",
            arg_type=str,
            arg_help="Environment to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="access-token",
            arg_type=str,
            arg_help="Access token to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="universe-id",
            arg_type=str,
            arg_help="Universe ID to use",
            mandatory=False,
        ),
    ]

    module_functions = [
        action_code,
        workspace,
        board,
        menu_path,
        component,
        activity,
        file,
        ai_output_file,
        role,
    ]

    for func in module_functions:
        get_parser.decor_add_command(common_args=common_args)(func)

    return get_parser


async def action_code(
    action: str = CLIFuncParam(prompt=True),
    save_to_path: str = CLIFuncParam(default=None, mandatory=False),
    **kwargs,
):
    """
    Get the code of an action
    :param action: Action name or id to get
    :param save_to_path: Path to save the code to, leave empty to print it
    """
    resource_getter = ResourceGetter(InitOptions(action=action, **kwargs))
    action_obj = await resource_getter.get_action()
    code = await action_obj.get_code()
    if save_to_path:
        with open(save_to_path, "w") as action_file:
            action_file.write(code)
        print(f"Code of the action ({action_obj['name']}) saved to {save_to_path}.")
    else:
        print(f"Code of the action ({action_obj['name']}):")
        print_code_with_line_numbers(code)


async def workspace(
    workspace_id: str = CLIFuncParam(prompt=True),
    show_theme: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Get a workspace
    :param workspace_id: UUID of the workspace to use
    :param show_theme: Flag to show the theme
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    business = await resource_getter.get_business()
    display_dict(
        business.cascade_to_dict(),
        fields=resource_getter.get_business_fields_to_show(show_theme=show_theme),
    )


async def board(
    workspace_id: Optional[str],
    board: str = CLIFuncParam(prompt=True),
    show_public_permission: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    show_theme: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Get a board
    :param workspace_id: UUID of the workspace to use
    :param board: Board name or id to get
    :param show_public_permission: Flag to show the public permission
    :param show_theme: Flag to show the theme
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, board=board, **kwargs)
    )
    dashboard = await resource_getter.get_dashboard()
    display_dict(
        dashboard.cascade_to_dict(),
        fields=resource_getter.get_dashboard_fields_to_show(
            show_public_permission=show_public_permission, show_theme=show_theme
        ),
    )


async def menu_path(
    workspace_id: Optional[str], menu_path: str = CLIFuncParam(prompt=True), **kwargs
):
    """Get a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path name or id to get
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    app = await resource_getter.get_app()
    display_dict(app.cascade_to_dict(), resource_getter.get_app_fields_to_show())


async def component(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(mandatory=False),
    component_id: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Get a component from a specific level, leave menu path and board empty to get a workspace component
    and specify only one to get a menu path or board component respectively. Don't specify both.
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path name or id to use
    :param component_id: Component name or id to get
    """
    resource_getter = ResourceGetter(
        InitOptions(
            workspace_id=workspace_id,
            menu_path=menu_path,
            component_id=component_id,
            **kwargs,
        )
    )
    report = await resource_getter.get_report()
    display_dict(report.cascade_to_dict(), resource_getter.get_report_fields_to_show())


async def activity(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(mandatory=False),
    activity: str = CLIFuncParam(prompt=True),
    show_template: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Get an activity from a specific level, leave menu path and board empty to get a workspace activity
    and specify only one to get a menu path or board activity respectively. Don't specify both.
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path name or id to use
    :param activity: Activity name or id to get
    :param show_template: Flag to show the template
    """
    resource_getter = ResourceGetter(
        InitOptions(
            workspace_id=workspace_id, menu_path=menu_path, activity=activity, **kwargs
        )
    )
    activity_obj = await resource_getter.get_activity()
    display_dict(
        activity_obj.cascade_to_dict(),
        resource_getter.get_activity_fields_to_show(show_template),
    )


async def file(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    file: str = CLIFuncParam(prompt=True),
    save_to_path: str = CLIFuncParam(default=None, mandatory=False),
    **kwargs,
):
    """Get a file object
    :param workspace_id: Workspace name or id to get
    :param menu_path: Menu path name or id to get
    :param file: File name or id to get
    :param save_to_path: Path to save the file to
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, file=file, **kwargs)
    )
    app = await resource_getter.get_app()
    file_obj = await resource_getter.get_file()
    if save_to_path:
        file_data = await app.get_file_object(file_obj["id"])
        save_as_file(logger, save_to_path, file_data, file_obj["name"])
        print(f"File ({file_obj['name']}) saved to {save_to_path}.")
    print(f"Metadata of the File ({file_obj['name']}):")
    display_dict(file_obj.cascade_to_dict(), ["tags", "metadata"])


async def ai_output_file(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    run_id: str = CLIFuncParam(prompt=True),
    file: str = CLIFuncParam(prompt=True),
    save_to_path: str = CLIFuncParam(default=None, mandatory=False),
    **kwargs,
):
    """Get an ai output file object
    :param workspace_id: Workspace name or id to get
    :param menu_path: Menu path name or id to get
    :param run_id: Run id to get the file from
    :param file: File name or id to get
    :param save_to_path: Path to save the file to
    """
    if not file:
        print("Please provide a file name or id.")
        return
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    ai_layer = await resource_getter.get_ai_layer()
    file_obj, metadata = await ai_layer.get_output_file_objects(
        run_id=run_id, file_name=file
    )
    if save_to_path:
        file_data = await ai_layer.get_output_file(file_obj["id"])
        save_as_file(logger, save_to_path, file_data, file_obj["name"])
        print(f"AI Output File ({metadata['name']}) saved to {save_to_path}.")
    print(f"Metadata of the File ({metadata['name']}):")
    display_dict(metadata, list(metadata.keys()))


async def role(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(group="menu_path_or_board", mandatory=False),
    board: str = CLIFuncParam(group="menu_path_or_board", mandatory=False),
    role: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Get a role from a specific level, leave menu path and board empty to get a workspace role
    and specify only one to get a menu path or board role respectively. Don't specify both.
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path name or id to use
    :param board: Board name or id to use
    :param role: Role name or id to get
    """
    resource_getter = ResourceGetter(
        InitOptions(
            workspace_id=workspace_id,
            menu_path=menu_path,
            board=board,
            role=role,
            **kwargs,
        )
    )
    role_obj = await resource_getter.get_role_dict()
    display_dict(role_obj, fields=list(role_obj.keys()))


if __name__ == "__main__":
    add_get_parser().parse_args()



import asyncio

from typing import Optional

from shimoku.api.resources.role import (
    user_delete_role,
    user_get_role,
    user_get_roles,
    user_create_role,
)
from shimoku.api.resources.business import Business
from shimoku.api.resources.app import App
from shimoku.api.resources.report import Report
from shimoku.api.resources.universe import Universe
from shimoku.utils import create_normalized_name
from shimoku.exceptions import WorkspaceError

import logging
from shimoku.execution_logger import log_error, ClassWithLogging

logger = logging.getLogger(__name__)


class WorkspacesLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the Workspace layer
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(
        self,
        universe: Universe,
    ):
        self._universe = universe
        self._get_for_roles = self._get_business_with_warning

    async def _check_similar_name(self, name: str):
        """
        Check if there is a workspace with a similar name
        :param name: name of the workspace
        """
        all_business = await self._universe.get_businesses()
        normalized_name = create_normalized_name(name)
        if normalized_name in [
            create_normalized_name(business["name"]) for business in all_business
        ]:
            log_error(
                logger,
                "There exists a workspace with a very similar name, please choose another name",
                WorkspaceError,
            )

    async def _get_business_with_warning(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[Business]:
        """
        Get a workspace with a warning if it does not exist
        :param name: name of the workspace
        :param uuid: uuid of the workspace
        """
        business: Business = await self._universe.get_business(uuid=uuid, name=name)
        if not business:
            logger.warning(f"Workspace ({name if name else uuid}) not found")
        return business

    async def create_workspace(
        self, name: str, create_default_roles: bool = True, theme: Optional[dict] = None
    ) -> dict:
        """Create a new workspace and the necessary roles if specified
        :param name: Name of the workspace
        :param create_default_roles: Create the default roles for the workspace
        :param theme: Theme of the workspace
        :return: workspace data
        """
        await self._check_similar_name(name)

        business = await self._universe.create_business(name=name, theme=theme)
        logger.info(f"Created workspace ({str(business)})")

        if create_default_roles:
            create_roles_tasks = []

            for role_permission_resource in [
                "DATA",
                "DATA_EXECUTION",
                "USER_MANAGEMENT",
                "BUSINESS_INFO",
            ]:
                create_roles_tasks.append(
                    business.create_role(
                        role="business_read",
                        resource=role_permission_resource,
                    )
                )

            await asyncio.gather(*create_roles_tasks)
            logger.info(f"Created default roles for workspace ({str(business)})")

        return business.cascade_to_dict()

    async def update_workspace(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        theme: Optional[dict] = None,
    ) -> bool:
        """Update a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param new_name: New name of the workspace
        :param theme: New theme of the workspace
        :return: True if the workspace was updated
        """
        params = {}

        if new_name is not None:
            await self._check_similar_name(new_name)
            params["new_name"] = new_name

        if isinstance(theme, dict):
            params["theme"] = theme

        return await self._universe.update_business(uuid=uuid, name=name, **params)

    async def delete_workspace(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        """Delete a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: True if the workspace was deleted
        """
        return await self._universe.delete_business(uuid=uuid, name=name)

    async def get_workspace(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> dict:
        """Get a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: Workspace data
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return {}
        return business.cascade_to_dict()

    async def get_workspace_menu_paths(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict]:
        """Get the menu paths of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of apps
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return []
        return [app.cascade_to_dict() for app in await business.get_apps()]

    async def get_workspace_menu_path_ids(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict]:
        """Get the menu path ids of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of app ids
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return []
        return [app["id"] for app in await business.get_apps()]

    async def get_workspace_boards(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict]:
        """Get the boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of dashboards
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return []
        return [
            dashboard.cascade_to_dict() for dashboard in await business.get_dashboards()
        ]

    async def delete_all_workspace_menu_paths(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """Delete all menu paths of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return
        apps = await business.get_apps()
        await asyncio.gather(*[business.delete_app(uuid=app["id"]) for app in apps])

    async def delete_all_workspace_boards(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        force: bool = False,
    ):
        """Delete all boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param force: Flag to force delete the boards
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return
        dashboards = await business.get_dashboards()
        if force:
            await asyncio.gather(
                *[dashboard.remove_all_apps() for dashboard in dashboards]
            )
        await asyncio.gather(
            *[
                business.delete_dashboard(uuid=dashboard["id"])
                for dashboard in dashboards
            ]
        )

    async def change_boards_order(
        self,
        boards: list[str],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Change the order of the boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param boards: list of dashboard names
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return
        await asyncio.gather(
            *[
                business.update_dashboard(name=d_name, order=i + 1)
                for i, d_name in enumerate(boards)
            ]
        )

    @staticmethod
    async def _change_sub_paths_order(
        business: Business, menu_path: str, sub_paths: list[str]
    ):
        """
        Change the order of the sub paths of a menu path
        :param business: Business object
        :param menu_path: Menu path name
        :param sub_paths: list of sub paths
        """
        app: App = await business.get_app(name=menu_path, create_if_not_exists=False)
        if not app:
            logger.warning(f"Menu path ({menu_path}) not found")
            return
        reports: list[Report] = await app.get_reports()
        sub_paths = [create_normalized_name(sub_path) for sub_path in sub_paths]
        non_referenced_reports = sorted(
            [
                report
                for report in reports
                if report["path"] is not None
                and create_normalized_name(report["path"]) not in sub_paths
            ],
            key=lambda x: x["pathOrder"] if x["pathOrder"] is not None else 0,
        )
        for report in non_referenced_reports:
            sub_path = report["path"]
            if sub_path not in sub_paths:
                sub_paths.append(sub_path)

        tasks = []
        for i, sub_path in enumerate(sub_paths):
            _reports = [
                report
                for report in reports
                if (
                    report["path"] is not None
                    and create_normalized_name(report["path"])
                    == create_normalized_name(sub_path)
                )
            ]
            for report in _reports:
                tasks.append(app.update_report(uuid=report["id"], pathOrder=i))

        logger.info(f"Updating {len(reports)} components from menu path ({str(app)})")

        await asyncio.gather(*tasks)

    async def change_menu_order(
        self, menu_order: list, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """Change the order of the menu of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param menu_order: list of menu names
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return
        tasks = []
        for i, menu_option in enumerate(menu_order):
            menu_name = menu_option
            if isinstance(menu_option, tuple):
                menu_name, sub_paths = menu_option
                tasks.append(
                    self._change_sub_paths_order(
                        business=business, menu_path=menu_name, sub_paths=sub_paths
                    )
                )
            tasks.append(business.update_app(name=menu_name, order=i))

        await asyncio.gather(*tasks)

    # Accounts management
    async def invite_user(
        self, email: str, uuid: Optional[str] = None, name: Optional[str] = None,
        roles: Optional[list[str]] = None, send_email: bool = True
    ) -> Optional[dict]:
        """Invite a user to the workspace
        :param email: Email of the user
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param roles: Roles of the user
        :param send_email: Send the invitation email
        :return: Invitation data
        """
        if not roles:
            roles = []
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return None
        return (await business.invite_user(email, roles, send_email)).cascade_to_dict()

    async def get_pending_invitations(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict]:
        """Get the pending invitations of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of invitations
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return []
        return [
            invitation.cascade_to_dict()
            for invitation in await business.get_pending_invitations()
        ]

    async def delete_pending_invitation(
        self,
        email: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """Delete a pending invitation of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param email: Email of the user
        :return: True if the invitation was deleted
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return False

        if await business.delete_pending_invitation(email=email):
            logger.info(f"Successfully deleted pending invitation ({email})")
            return True
        else:
            logger.warning(f"Failed to delete pending invitation ({email})")
            return False

    async def get_all_workspace_users(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict]:
        """Get the users of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of users
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return []
        return [user.cascade_to_dict() for user in await business.get_users()]

    async def remove_user_from_workspace(
        self,
        email: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """Remove a user from a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param email: Email of the user
        :return: True if the user was removed
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return False

        if await business.delete_user(email=email):
            logger.info(f"Successfully removed user ({email} from workspace")
        else:
            logger.warning(f"Failed to remove user ({email}) from workspace")

    # Role management
    get_role = user_get_role
    get_roles = user_get_roles
    create_role = user_create_role
    delete_role = user_delete_role



from shimoku import Client


def predictive_line(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "predictive-line-test")
    shimoku_client.plt.predictive_line(
        data="main data",
        x="date",
        min_value_mark=3,
        max_value_mark=4,
        order=1,
        rows_size=2,
        cols_size=12,
    )



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame
from dataclasses import dataclass
import shimoku


class WorkspacesLayerHeader:
    """
    This class is used to interact with the API at the Workspace layer
    """

    def change_boards_order(
        self,
        boards: list,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Change the order of the boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param boards: list of dashboard names
        """
        pass

    def change_menu_order(
        self,
        menu_order: list,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Change the order of the menu of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param menu_order: list of menu names
        """
        pass

    def create_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        resource: Optional[str] = None,
        role_name: Optional[str] = None,
        permission: Optional[str] = None,
        target: Optional[str] = None,
    ) -> dict:
        """
        Create a new role at the resource level
        :param name: name of the resource
        :param uuid: uuid of the resource
        :param resource: resource of the role
        :param role_name: name of the role
        :param permission: permission of the role
        :param target: target of the role
        """
        pass

    def create_workspace(
        self,
        name: str,
        create_default_roles: bool = True,
        theme: Optional[dict] = None,
    ) -> dict:
        """Create a new workspace and the necessary roles if specified
        :param name: Name of the workspace
        :param create_default_roles: Create the default roles for the workspace
        :param theme: Theme of the workspace
        :return: workspace data
        """
        pass

    def delete_all_workspace_boards(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        force: bool = False,
    ):
        """Delete all boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param force: Flag to force delete the boards
        """
        pass

    def delete_all_workspace_menu_paths(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Delete all menu paths of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        """
        pass

    def delete_pending_invitation(
        self,
        email: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """Delete a pending invitation of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param email: Email of the user
        :return: True if the invitation was deleted
        """
        pass

    def delete_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_id: Optional[str] = None,
        role_name: Optional[str] = None,
    ):
        """
        Delete the role at the resource level
        :param name: name of the resource
        :param uuid: uuid of the resource
        :param role_id: id of the role
        :param role_name: name of the role
        """
        pass

    def delete_workspace(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """Delete a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: True if the workspace was deleted
        """
        pass

    def get_all_workspace_users(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """Get the users of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of users
        """
        pass

    def get_pending_invitations(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """Get the pending invitations of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of invitations
        """
        pass

    def get_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_name: Optional[str] = None,
    ) -> list:
        """
        Get the role at the resource level
        :param name: name of the resource
        :param uuid: uuid of the resource
        :param role_name: name of the role
        """
        pass

    def get_roles(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the roles at the resource level
        :param name: name of the resource
        :param uuid: uuid of the resource
        """
        pass

    def get_workspace(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """Get a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: Workspace data
        """
        pass

    def get_workspace_boards(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """Get the boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of dashboards
        """
        pass

    def get_workspace_menu_path_ids(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """Get the menu path ids of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of app ids
        """
        pass

    def get_workspace_menu_paths(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """Get the menu paths of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of apps
        """
        pass

    def invite_user(
        self,
        email: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        roles: Optional[list[str]] = None,
        send_email: bool = True,
    ) -> Optional[dict]:
        """Invite a user to the workspace
        :param email: Email of the user
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param roles: Roles of the user
        :param send_email: Send the invitation email
        :return: Invitation data
        """
        pass

    def remove_user_from_workspace(
        self,
        email: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """Remove a user from a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param email: Email of the user
        :return: True if the user was removed
        """
        pass

    def update_workspace(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        theme: Optional[dict] = None,
    ) -> bool:
        """Update a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param new_name: New name of the workspace
        :param theme: New theme of the workspace
        :return: True if the workspace was updated
        """
        pass



indicator_color_by_value_data = [
    {
        "align": "center",
        "title": "Positive case",
        "value": 10,
    },
    {
        "align": "center",
        "title": "Negative case",
        "value": -10,
    },
    {
        "align": "center",
        "title": "Neutral case",
        "value": 0,
    },
]



from typing import Union, List
from pydantic import BaseModel
import strawberry


@strawberry.type
class CreateUniverseApiKeyInput(BaseModel):
    description: str
    id: Union[str, None] = None
    universeId: Union[str, None] = None
    userType: Union[str, None] = None


@strawberry.type
class DeleteUniverseApiKeyInput(BaseModel):
    id: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class UpdateBusinessInput(BaseModel):
    id: Union[str, None] = None
    name: Union[str, None] = None
    type: Union[str, None] = None
    createdAt: Union[str, None] = None
    etlMachineLastProcessed: Union[str, None] = None
    integrationId: Union[str, None] = None
    defaultstr: Union[str, None] = None
    businessUniverseId: Union[str, None] = None
    theme: Union[str, None] = None


@strawberry.type
class CreateAppInput(BaseModel):
    name: str
    normalizedName: str
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    order: Union[int, None] = None
    hidePath: Union[bool, None] = None
    showBreadcrumb: Union[bool, None] = None
    showHistoryNavigation: Union[bool, None] = None
    hideTitle: Union[bool, None] = None
    isDisabled: Union[bool, None] = None
    appBusinessId: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class UpdateAppInput(BaseModel):
    id: Union[str, None] = None
    name: Union[str, None] = None
    normalizedName: Union[str, None] = None
    createdAt: Union[str, None] = None
    order: Union[int, None] = None
    hidePath: Union[bool, None] = None
    showBreadcrumb: Union[bool, None] = None
    showHistoryNavigation: Union[bool, None] = None
    hideTitle: Union[bool, None] = None
    isDisabled: Union[bool, None] = None
    appBusinessId: Union[str, None] = None
    universeId: Union[str, None] = None
    paymentType: Union[str, None] = None
    trialDays: Union[int, None] = None
    appSubscriptionInUseId: Union[str, None] = None
    accountId: Union[str, None] = None
    integrationId: Union[str, None] = None
    churnedDate: Union[str, None] = None
    appTypeId: Union[str, None] = None


@strawberry.type
class DeleteAppInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class PublicPermissionInput(BaseModel):
    isPublic: bool
    permission: str
    token: str


@strawberry.type
class CreateDashboardInput(BaseModel):
    name: str
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    businessId: Union[str, None] = None
    publicPermission: Union[PublicPermissionInput, None] = None
    order: Union[int, None] = None
    isDisabled: Union[bool, None] = None
    theme: Union[str, None] = None


@strawberry.type
class UpdateDashboardInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    businessId: Union[str, None] = None
    publicPermission: Union[PublicPermissionInput, None] = None
    name: Union[str, None] = None
    order: Union[int, None] = None
    isDisabled: Union[bool, None] = None
    universeId: Union[str, None] = None
    theme: Union[str, None] = None


@strawberry.type
class DeleteDashboardInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateAppDashboardInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    dashboardId: Union[str, None] = None
    appId: Union[str, None] = None


@strawberry.type
class UpdateAppDashboardInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    dashboardId: Union[str, None] = None
    appId: Union[str, None] = None


@strawberry.type
class DeleteAppDashboardInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateActivityWebHookInput(BaseModel):
    url: str
    method: str
    id: Union[str, None] = None
    headers: Union[str, None] = None
    activityId: Union[str, None] = None


@strawberry.type
class UpdateActivityWebHookInput(BaseModel):
    id: Union[str, None] = None
    url: Union[str, None] = None
    method: Union[str, None] = None
    headers: Union[str, None] = None
    activityId: Union[str, None] = None


@strawberry.type
class DeleteActivityWebHookInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateActivityInput(BaseModel):
    name: str
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    appId: Union[str, None] = None
    settings: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class UpdateActivityInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    appId: Union[str, None] = None
    name: Union[str, None] = None
    settings: Union[str, None] = None
    activityWebHookId: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class DeleteActivityInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateActionInput(BaseModel):
    id: Union[str, None] = None
    name: str
    description: Union[str, None] = None
    actionTemplateId: Union[str, None] = None
    universeApiKeyId: Union[str, None] = None


@strawberry.type
class UpdateActionInput(BaseModel):
    id: Union[str, None] = None
    name: Union[str, None] = None
    description: Union[str, None] = None
    universeApiKeyId: Union[str, None] = None


@strawberry.type
class DeleteActionInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateActionScriptInput(BaseModel):
    order: int
    codeFragment: str
    id: Union[str, None] = None
    actionId: Union[str, None] = None


@strawberry.type
class UpdateActionScriptInput(BaseModel):
    id: Union[str, None] = None
    order: Union[int, None] = None
    codeFragment: Union[str, None] = None
    actionId: Union[str, None] = None


@strawberry.type
class DeleteActionScriptInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateRunInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    activityId: Union[str, None] = None
    settings: Union[str, None] = None
    isWebhookCalled: Union[bool, None] = None


@strawberry.type
class UpdateRunInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    activityId: Union[str, None] = None
    settings: Union[str, None] = None
    isWebhookCalled: Union[bool, None] = None


@strawberry.type
class DeleteRunInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateLogInput(BaseModel):
    id: Union[str, None] = None
    message: Union[str, None] = None
    runId: Union[str, None] = None
    dateTime: Union[str, None] = None
    tags: Union[str, None] = None
    severity: Union[str, None] = None


@strawberry.type
class UpdateLogInput(BaseModel):
    id: Union[str, None] = None
    message: Union[str, None] = None
    runId: Union[str, None] = None
    dateTime: Union[str, None] = None
    tags: Union[str, None] = None
    severity: Union[str, None] = None


@strawberry.type
class DeleteLogInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateReportInput(BaseModel):
    order: int
    id: Union[str, None] = None
    appId: Union[str, None] = None
    createdAt: Union[str, None] = None
    grid: Union[str, None] = None
    sizeRows: Union[int, None] = None
    sizeColumns: Union[int, None] = None
    sizePadding: Union[str, None] = None
    isDisabled: Union[bool, None] = None
    reportType: Union[str, None] = None
    path: Union[str, None] = None
    pathOrder: Union[int, None] = None
    dataFields: Union[str, None] = None
    title: Union[str, None] = None
    description: Union[str, None] = None
    chartData: Union[str, None] = None
    chartDataItem: Union[str, None] = None
    chartDataAux: Union[str, None] = None
    smartFilters: Union[str, None] = None
    subscribe: Union[bool, None] = None
    properties: Union[str, None] = None
    bentobox: Union[str, None] = None
    hidePath: Union[bool, None] = None
    showBreadcrumb: Union[bool, None] = None
    showHistoryNavigation: Union[bool, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class UpdateReportInput(BaseModel):
    id: Union[str, None] = None
    appId: Union[str, None] = None
    createdAt: Union[str, None] = None
    order: Union[int, None] = None
    grid: Union[str, None] = None
    sizeRows: Union[int, None] = None
    sizeColumns: Union[int, None] = None
    sizePadding: Union[str, None] = None
    isDisabled: Union[bool, None] = None
    reportType: Union[str, None] = None
    path: Union[str, None] = None
    pathOrder: Union[int, None] = None
    dataFields: Union[str, None] = None
    title: Union[str, None] = None
    description: Union[str, None] = None
    chartData: Union[str, None] = None
    chartDataItem: Union[str, None] = None
    chartDataAux: Union[str, None] = None
    smartFilters: Union[str, None] = None
    subscribe: Union[bool, None] = None
    properties: Union[str, None] = None
    bentobox: Union[str, None] = None
    hidePath: Union[bool, None] = None
    showBreadcrumb: Union[bool, None] = None
    showHistoryNavigation: Union[bool, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class DeleteReportInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateReportDataSetInput(BaseModel):
    id: Union[str, None] = None
    dataSetId: Union[str, None] = None
    reportId: Union[str, None] = None
    properties: Union[str, None] = None


@strawberry.type
class UpdateReportDataSetInput(BaseModel):
    id: Union[str, None] = None
    dataSetId: Union[str, None] = None
    reportId: Union[str, None] = None
    properties: Union[str, None] = None


@strawberry.type
class DeleteReportDataSetInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateDataSetInput(BaseModel):
    id: Union[str, None] = None
    name: Union[str, None] = None
    dataSetAppId: Union[str, None] = None
    universeId: Union[str, None] = None
    columns: Union[str, None] = None


@strawberry.type
class UpdateDataSetInput(BaseModel):
    id: Union[str, None] = None
    name: Union[str, None] = None
    columns: Union[str, None] = None
    dataSetAppId: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class DeleteDataSetInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateDataInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    dataSetId: Union[str, None] = None
    description: Union[str, None] = None
    stringField1: Union[str, None] = None
    stringField2: Union[str, None] = None
    stringField3: Union[str, None] = None
    stringField4: Union[str, None] = None
    stringField5: Union[str, None] = None
    stringField6: Union[str, None] = None
    stringField7: Union[str, None] = None
    stringField8: Union[str, None] = None
    orderField1: Union[float, None] = None
    intField1: Union[float, None] = None
    intField2: Union[float, None] = None
    intField3: Union[float, None] = None
    intField4: Union[float, None] = None
    intField5: Union[float, None] = None
    intField6: Union[float, None] = None
    intField7: Union[float, None] = None
    intField8: Union[float, None] = None
    dateField1: Union[str, None] = None
    dateField2: Union[str, None] = None
    dateField3: Union[str, None] = None
    dateField4: Union[str, None] = None
    dateField5: Union[str, None] = None
    customField1: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class BatchCreateDataInput(BaseModel):
    data: List[CreateDataInput]


@strawberry.type
class UpdateDataInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    dataSetId: Union[str, None] = None
    description: Union[str, None] = None
    stringField1: Union[str, None] = None
    stringField2: Union[str, None] = None
    stringField3: Union[str, None] = None
    stringField4: Union[str, None] = None
    stringField5: Union[str, None] = None
    stringField6: Union[str, None] = None
    stringField7: Union[str, None] = None
    stringField8: Union[str, None] = None
    orderField1: Union[float, None] = None
    intField1: Union[float, None] = None
    intField2: Union[float, None] = None
    intField3: Union[float, None] = None
    intField4: Union[float, None] = None
    intField5: Union[float, None] = None
    intField6: Union[float, None] = None
    intField7: Union[float, None] = None
    intField8: Union[float, None] = None
    dateField1: Union[str, None] = None
    dateField2: Union[str, None] = None
    dateField3: Union[str, None] = None
    dateField4: Union[str, None] = None
    dateField5: Union[str, None] = None
    customField1: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class DeleteDataInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateReportEntryInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    reportId: Union[str, None] = None
    description: Union[str, None] = None
    stringField1: Union[str, None] = None
    stringField2: Union[str, None] = None
    stringField3: Union[str, None] = None
    stringField4: Union[str, None] = None
    intField1: Union[float, None] = None
    intField2: Union[float, None] = None
    intField3: Union[float, None] = None
    intField4: Union[float, None] = None
    data: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class UpdateReportEntryInput(BaseModel):
    id: Union[str, None] = None
    createdAt: Union[str, None] = None
    reportId: Union[str, None] = None
    description: Union[str, None] = None
    stringField1: Union[str, None] = None
    stringField2: Union[str, None] = None
    stringField3: Union[str, None] = None
    stringField4: Union[str, None] = None
    intField1: Union[float, None] = None
    intField2: Union[float, None] = None
    intField3: Union[float, None] = None
    intField4: Union[float, None] = None
    data: Union[str, None] = None
    universeId: Union[str, None] = None


@strawberry.type
class DeleteReportEntryInput(BaseModel):
    id: Union[str, None] = None


@strawberry.type
class CreateFileInput(BaseModel):
    id: Union[str, None] = None
    appId: Union[str, None] = None
    fileName: Union[str, None] = None
    name: Union[str, None] = None
    universeId: Union[str, None] = None
    contentType: Union[str, None] = None


@strawberry.type
class UpdateFileInput(BaseModel):
    id: Union[str, None] = None
    appId: Union[str, None] = None
    fileName: Union[str, None] = None
    name: Union[str, None] = None
    universeId: Union[str, None] = None
    contentType: Union[str, None] = None


@strawberry.type
class DeleteFileInput(BaseModel):
    id: Union[str, None] = None
